from .constraint import *  # noqa F403
from .node import *  # noqa F403
from .edge import *  # noqa F403
from .network import *  # noqa F403


__all__ = [name for name in dir() if not name.startswith('_')]



import numpy as np

import jax.numpy as jnp

from jax_fdm.constraints.node import NodeConstraint

from jax_fdm.geometry import curvature_point_polygon


class NodeCurvatureConstraint(NodeConstraint):
    """
    Constraints the (discrete) curvature of a node based on its surrounding polygon of neighboring nodes.
    """
    def __init__(self, key, polygon, bound_low, bound_up):
        super().__init__(key, bound_low, bound_up)
        self.polygon = polygon
        self.index_polygon = None

    def init(self, model, structure):
        """
        Initialize the constraint with information from an equilibrium model.
        """
        super().init(model, structure)
        self.index_polygon = self.polygon_indices(model, structure)

    def polygon_indices(self, model, structure):
        """
        Obtains the indices of the polygon from a model.
        """
        index_max = max(self.index) + 1
        polygon = np.atleast_2d(self.polygon)
        index_polygon = np.zeros((index_max, polygon.shape[1]))
        for p, idx in zip(polygon, self.index):
            index_polygon[idx, :] = tuple([structure.node_index[nbr] for nbr in p])

        return jnp.array(index_polygon, dtype=jnp.int64)

    def constraint(self, eqstate, index):
        """
        Returns the curvature at a node based on the xyz coordinates of its one-hop neighborhood.
        """
        point = eqstate.xyz[index, :]
        index_polygon = self.index_polygon[index, :]
        polygon = eqstate.xyz[index_polygon, :]

        return curvature_point_polygon(point, polygon)



from equinox import is_array

import jax.numpy as jnp

from jax.experimental.sparse import CSC

from jax_fdm.equilibrium.states import EquilibriumState

from jax_fdm.equilibrium.sparse import sparse_solve as spsolve

from jax_fdm.equilibrium.solvers import fixed_point
from jax_fdm.equilibrium.solvers import solver_forward

from jax_fdm.equilibrium.loads import nodes_load_from_faces
from jax_fdm.equilibrium.loads import nodes_load_from_edges

from jax_fdm.equilibrium.states import LoadState


# ==========================================================================
# Equilibrium model
# ==========================================================================

class EquilibriumModel:
    """
    The equilibrium model.
    """
    def __init__(self,
                 tmax=100,
                 eta=1e-6,
                 is_load_local=False,
                 itersolve_fn=None,
                 implicit_diff=True,
                 load_nodes_iter=False,
                 verbose=False):

        self.tmax = tmax
        self.eta = eta
        self.is_load_local = is_load_local
        self.linearsolve_fn = jnp.linalg.solve
        self.itersolve_fn = itersolve_fn or solver_forward
        self.implicit_diff = implicit_diff
        self.load_nodes_iter = load_nodes_iter
        self.verbose = verbose

    # ----------------------------------------------------------------------
    # Edges
    # ----------------------------------------------------------------------

    @staticmethod
    def edges_vectors(xyz, connectivity):
        """
        Calculate the unnormalized edge directions (nodal coordinate differences).
        """
        return connectivity @ xyz

    @staticmethod
    def edges_lengths(vectors):
        """
        Compute the length of the edges.
        """
        return jnp.linalg.norm(vectors, axis=1, keepdims=True)

    @staticmethod
    def edges_forces(q, lengths):
        """
        Calculate the force in the edges.
        """
        return jnp.reshape(q, (-1, 1)) * lengths

    # ----------------------------------------------------------------------
    # Nodes
    # ----------------------------------------------------------------------

    @staticmethod
    def nodes_residuals(q, loads, vectors, connectivity):
        """
        Compute the residual forces on the nodes of the structure.
        """
        return loads - connectivity.T @ (q[:, None] * vectors)

    @staticmethod
    def nodes_positions(xyz_free, xyz_fixed, indices):
        """
        Concatenate in order the position of the free and the fixed nodes.
        """
        return jnp.concatenate((xyz_free, xyz_fixed))[indices, :]

    def nodes_free_positions(self, q, xyz_fixed, loads, structure):
        """
        Calculate the XYZ coordinates of the free nodes.
        """
        A = self.stiffness_matrix(q, structure)
        b = self.force_matrix(q, xyz_fixed, loads, structure)

        return self.linearsolve_fn(A, b)

    def nodes_equilibrium(self, q, xyz_fixed, loads_nodes, structure):
        """
        Calculate static equilibrium on the nodes of a structure.
        """
        indices = structure.indices_freefixed
        xyz_free = self.nodes_free_positions(q, xyz_fixed, loads_nodes, structure)

        return self.nodes_positions(xyz_free, xyz_fixed, indices)

    # --------------------------------------------------------------------
    #  Load it up!
    # --------------------------------------------------------------------

    def nodes_load(self, xyz, loads, structure):
        """
        Calculate the loads applied to the nodes of the structure.
        """
        nodes_load, edges_load, faces_load = loads

        if is_array(edges_load):
            if edges_load.size > 1:
                edges_load_ = self.edges_load(xyz, edges_load, structure, self.is_load_local)
                nodes_load = nodes_load + edges_load_

        if is_array(faces_load):
            if faces_load.size > 1:
                faces_load_ = self.faces_load(xyz, faces_load, structure, self.is_load_local)
                nodes_load = nodes_load + faces_load_

        return nodes_load

    @staticmethod
    def faces_load(xyz, faces_load, structure, is_local=False):
        """
        Calculate the tributary face loads aplied to the nodes of a structure.
        """
        return nodes_load_from_faces(xyz, faces_load, structure, is_local)

    @staticmethod
    def edges_load(xyz, edges_load, structure, is_local=False):
        """
        Calculate the tributary edge loads aplied to the nodes of a structure.
        """
        return nodes_load_from_edges(xyz, edges_load, structure, is_local)

    # ------------------------------------------------------------------------------
    #  Call me, maybe
    # ------------------------------------------------------------------------------

    def __call__(self, params, structure):
        """
        Compute an equilibrium state using the force density method (FDM).
        """
        q, xyz_fixed, loads_state = params
        loads_nodes = loads_state.nodes

        tmax = self.tmax
        eta = self.eta
        solver = self.itersolve_fn
        implicit_diff = self.implicit_diff
        verbose = self.verbose

        xyz = self.equilibrium(q, xyz_fixed, loads_nodes, structure)

        if tmax > 1:
            # Setting node loads to zero when tmax > 1 if specified
            if not self.load_nodes_iter:
                loads_nodes = jnp.zeros_like(loads_nodes)
                loads_state = LoadState(loads_nodes,
                                        loads_state.edges,
                                        loads_state.faces)

            xyz = self.equilibrium_iterative(q,
                                             xyz_fixed,
                                             loads_state,
                                             structure,
                                             xyz_init=xyz,
                                             tmax=tmax,
                                             eta=eta,
                                             solver=solver,
                                             implicit_diff=implicit_diff,
                                             verbose=verbose)

        loads_nodes = self.nodes_load(xyz, loads_state, structure)

        return self.equilibrium_state(q, xyz, loads_nodes, structure)

    # ------------------------------------------------------------------------------
    #  Equilibrium modes
    # ------------------------------------------------------------------------------

    def equilibrium(self, q, xyz_fixed, loads_nodes, structure):
        """
        Calculate static equilibrium on a structure.
        """
        return self.nodes_equilibrium(q, xyz_fixed, loads_nodes, structure)

    def equilibrium_iterative(self,
                              q,
                              xyz_fixed,
                              load_state,
                              structure,
                              xyz_init=None,
                              tmax=100,
                              eta=1e-6,
                              solver=None,
                              implicit_diff=True,
                              verbose=False):
        """
        Calculate static equilibrium on a structure iteratively.

        Notes
        -----
        This function only supports reverse mode auto-differentiation.
        To support forward-mode, we should define a custom jvp using implicit differentiation.
        """
        def equilibrium_iterative_fn(params, xyz_init):
            """
            This closure function avoids re-computing A and f_fixed throughout iterations
            because these two matrices remain constant during the fixed point search.

            TODO: Extract closure into function shared with the other nodes equilibrium function?
            """
            A, f_fixed, xyz_fixed, load_state = params

            free = structure.indices_free
            freefixed = structure.indices_freefixed

            loads_nodes = self.nodes_load(xyz_init, load_state, structure)
            b = loads_nodes[free, :] - f_fixed
            xyz_free = self.linearsolve_fn(A, b)
            xyz_ = self.nodes_positions(xyz_free, xyz_fixed, freefixed)

            return xyz_

        # recompute xyz_init if not input
        if xyz_init is None:
            xyz_init = self.equilibrium(q, xyz_fixed, load_state.nodes, structure)

        A = self.stiffness_matrix(q, structure)
        f_fixed = self.force_fixed_matrix(q, xyz_fixed, structure)

        solver = solver or self.itersolve_fn
        solver_config = {"tmax": tmax,
                         "eta": eta,
                         "verbose": verbose,
                         "implicit": True}
        # For jaxopt compatibility (as it does not support sparse matrices yet)
        #  "implicit": False if self.linearsolve_fn is spsolve else True}

        solver_kwargs = {"solver_config": solver_config,
                         "f": equilibrium_iterative_fn,
                         "a": (A, f_fixed, xyz_fixed, load_state),
                         "x_init": xyz_init}

        if implicit_diff:
            # xyz_new = fixed_point(solver, **solver_kwargs)
            return fixed_point(solver, **solver_kwargs)

        # xyz_new = solver(**solver_kwargs)
        return solver(**solver_kwargs)

        # return xyz_new

    # ----------------------------------------------------------------------
    # Equilibrium state
    # ----------------------------------------------------------------------

    def equilibrium_state(self, q, xyz, loads_nodes, structure):
        """
        Assembles an equilibrium state object.
        """
        connectivity = structure.connectivity

        vectors = self.edges_vectors(xyz, connectivity)
        lengths = self.edges_lengths(vectors)
        residuals = self.nodes_residuals(q, loads_nodes, vectors, connectivity)
        forces = self.edges_forces(q, lengths)

        return EquilibriumState(xyz=xyz,
                                residuals=residuals,
                                lengths=lengths,
                                forces=forces,
                                loads=loads_nodes,
                                vectors=vectors)

    # ----------------------------------------------------------------------
    # Matrices
    # ----------------------------------------------------------------------

    @staticmethod
    def stiffness_matrix(q, structure):
        """
        The stiffness matrix of the structure.
        """
        c_free = structure.connectivity_free

        return c_free.T @ (q[:, None] * c_free)

    def force_matrix(self, q, xyz_fixed, loads, structure):
        """
        The force residual matrix of the structure.
        """
        free = structure.indices_free

        return loads[free, :] - self.force_fixed_matrix(q, xyz_fixed, structure)

    @staticmethod
    def force_fixed_matrix(q, xyz_fixed, structure):
        """
        The force matrix block of the residual forces at the fixed nodes.
        """
        c_free = structure.connectivity_free
        c_fixed = structure.connectivity_fixed

        return c_free.T @ (q[:, None] * (c_fixed @ xyz_fixed))


# ==========================================================================
# Sparse equilibrium model
# ==========================================================================

class EquilibriumModelSparse(EquilibriumModel):
    """
    The equilibrium model. Sparse.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.linearsolve_fn = spsolve

    @staticmethod
    def stiffness_matrix(q, structure):
        """
        Computes the LHS matrix in CSC format from a vector of force densities.
        """
        # shorthands
        index_array = structure.index_array
        diag_indices = structure.diag_indices
        diags = structure.diags

        nondiags_data = -q[index_array.data - 1]
        args = (nondiags_data, index_array.indices, index_array.indptr)

        nondiags = CSC(args, shape=index_array.shape)

        # sum of force densities for each node
        diag_fd = diags.T @ q

        nondiags.data = nondiags.data.at[diag_indices].set(diag_fd)

        return nondiags



import jax.numpy as jnp

from jax_fdm.goals import ScalarGoal
from jax_fdm.goals.edge import EdgeGoal


class EdgeLoadPathGoal(ScalarGoal, EdgeGoal):
    """
    Make an edge of a network to reach a target force.
    """
    @staticmethod
    def prediction(eq_state, index):
        """
        The predicted edge force.
        """
        return jnp.abs(eq_state.forces[index]) * eq_state.lengths[index]



import jax.numpy as jnp

from jax_fdm.goals import ScalarGoal
from jax_fdm.goals.edge import EdgeGoal


class EdgeLengthGoal(ScalarGoal, EdgeGoal):
    """
    Make an edge of a network to reach a certain length.
    """
    @staticmethod
    def prediction(eq_state, index):
        """
        The current edge length.
        """
        return eq_state.lengths[index]


class EdgesLengthEqualGoal(ScalarGoal, EdgeGoal):
    """
    Equalize the length of a selection of edges by minimizing
    the normalized variance of their lengths.
    """
    def __init__(self, key, weight=1.0):
        super().__init__(key=key, target=0.0, weight=weight)
        self.is_collectible = False

    def init(self, model, structure):
        """
        Initialize the goal with information from an equilibrium model.
        """
        self.index = [super().index_from_model(model, structure)]

    @staticmethod
    def prediction(eq_state, index):
        """
        The normalized variance of the lengths of the edges.
        """
        lengths = eq_state.lengths[index]

        return jnp.atleast_1d(jnp.var(lengths) / jnp.mean(lengths))



import numpy as np

import jax.numpy as jnp

from jax_fdm.geometry import angle_vectors

from jax_fdm.goals import ScalarGoal
from jax_fdm.goals.edge import EdgeGoal


class EdgeAngleGoal(ScalarGoal, EdgeGoal):
    """
    Reach a target angle between the direction of an edge and a reference vector.
    """
    def __init__(self, key, vector, target, weight=1.0):
        super().__init__(key=key, target=target, weight=weight)
        self._vector = None
        self.vector = vector

    @property
    def vector(self):
        """
        The vector to take the angle with.
        """
        return self._vector

    @vector.setter
    def vector(self, vector):
        self._vector = jnp.reshape(jnp.asarray(vector), (-1, 3))

    def init(self, model, structure):
        """
        Initialize the goal with information from an equilibrium model.
        """
        super().init(model, structure)

        # create matrix of vectors
        vector = self.vector
        vm = np.zeros((max(self.index) + 1, 3))
        for v, idx in zip(vector, self.index):
            vm[idx, :] = v
        self.vector = vm

    def prediction(self, eq_state, index):
        """
        The angle between the edge and the reference vector.
        """
        vector = eq_state.vectors[index, :]
        return angle_vectors(vector, self.vector[index, :])



from jax_fdm.constraints.node import NodeConstraint


class NodeXCoordinateConstraint(NodeConstraint):
    """
    Constraint the X coordinate of a node between a lower and an upper bound.
    """
    @staticmethod
    def constraint(eqstate, index):
        """
        Returns the X coordinate of a node from an equilibrium state.
        """
        return eqstate.xyz[index, :1]


class NodeYCoordinateConstraint(NodeConstraint):
    """
    Constraint the Y coordinate of a node between a lower and an upper bound.
    """
    @staticmethod
    def constraint(eqstate, index):
        """
        Returns the Y coordinate of a node from an equilibrium state.
        """
        return eqstate.xyz[index, 1:2]


class NodeZCoordinateConstraint(NodeConstraint):
    """
    Constraint the Z coordinate of a node between a lower and an upper bound.
    """
    @staticmethod
    def constraint(eqstate, index):
        """
        Returns the Z coordinate of a node from an equilibrium state.
        """
        return eqstate.xyz[index, 2]



from jax_fdm.losses import Error
from jax_fdm.losses import Regularizer


# ==========================================================================
# Loss
# ==========================================================================

class Loss:
    """
    A function composed of error and regularization terms.
    """
    def __init__(self, *args, name=None):
        self._terms_error = None
        self._terms_regularization = None

        self.terms_error = args
        self.terms_regularization = args
        self.name = name or self.__class__.__name__

    def __call__(self, params, model, structure):
        """
        Compute the scalar output of the loss function.
        """
        eq_state = model(params, structure)

        loss = 0.0
        for error_term in self.terms_error:
            loss = loss + error_term(eq_state)

        for reg_term in self.terms_regularization:
            loss = loss + reg_term(params)

        return loss

    @property
    def terms_error(self):
        """
        The error terms in the loss function.
        """
        return self._error_terms

    @terms_error.setter
    def terms_error(self, terms):
        self._error_terms = [term for term in terms if isinstance(term, Error)]

    @property
    def terms_regularization(self):
        """
        The regularization terms in the loss function.
        """
        return self._regularization_terms

    @terms_regularization.setter
    def terms_regularization(self, terms):
        self._regularization_terms = [term for term in terms if isinstance(term, Regularizer)]

    @property
    def terms(self):
        """
        The error and regularization terms of the loss function.
        """
        return self.terms_error + self.terms_regularization

    def number_of_goals(self):
        """
        The total number of individual goals for all error terms in the loss.
        """
        return sum(term.number_of_goals() for term in self.terms_error)

    def number_of_regularizers(self):
        """
        The total number of regularization terms in the loss.
        """
        return len(self.terms_regularization)

    def number_of_collections(self):
        """
        The total number of goal collections for all error terms in the loss.
        """
        return sum(term.number_of_collections() for term in self.terms_error)



"""
A gradient-based optimizer that deals with equality and inequality constraints.
"""
from itertools import groupby

from functools import partial

from jax import jit
from jax import jacfwd

from scipy.optimize import NonlinearConstraint

from jax_fdm.optimization import Collection
from jax_fdm.optimization.optimizers import Optimizer


# ==========================================================================
# Constrained optimizer
# ========== ===============================================================

class ConstrainedOptimizer(Optimizer):
    """
    A gradient-based optimizer that handles constraints.
    """
    def constraints(self, constraints, model, structure, params_opt):
        """
        Returns the defined constraints in a format amenable to `scipy.minimize`.
        """
        if not constraints:
            return

        print(f"Constraints: {len(constraints)}")
        constraints = self.collect_constraints(constraints)
        print(f"\tConstraint colections: {len(constraints)}")

        clist = []
        for constraint in constraints:

            # initialize constraint
            constraint.init(model, structure)

            # gather information for scipy constraint
            fun = partial(self.constraint,
                          constraint=constraint,
                          model=model,
                          structure=structure)

            fun = jit(fun)
            jac = jit(jacfwd(fun))

            lb = constraint.bound_low
            ub = constraint.bound_up

            # warm start
            fun(params_opt)
            jac(params_opt)

            # store non linear constraint
            clist.append(NonlinearConstraint(fun=fun, jac=jac, lb=lb, ub=ub))

        return clist

    def constraint(self, params_opt, constraint, model, structure):
        """
        A wrapper around a constraint callable object.
        """
        params = self.parameters_fdm(params_opt)

        return constraint(params, model, structure)

    @staticmethod
    def collect_constraints(constraints):
        """
        Convert a list of constraints into a list of constraint collections.
        """
        constraints = sorted(constraints, key=lambda g: type(g).__name__)
        groups = groupby(constraints, lambda g: type(g))

        collections = []
        for _, group in groups:
            collection = Collection(list(group))
            collections.append(collection)

        return collections



import numpy as np

from jax_fdm import DTYPE_JAX

from jax_fdm.datastructures import FDNetwork
from jax_fdm.datastructures import FDMesh

from jax_fdm.equilibrium import EquilibriumModel
from jax_fdm.equilibrium import EquilibriumModelSparse

from jax_fdm.equilibrium import EquilibriumStructure
from jax_fdm.equilibrium import EquilibriumMeshStructure
from jax_fdm.equilibrium import EquilibriumStructureSparse
from jax_fdm.equilibrium import EquilibriumMeshStructureSparse

from jax_fdm.equilibrium import EquilibriumParametersState


# ==========================================================================
# Form-finding
# ==========================================================================

def _fdm(model, params, structure, datastructure):
    """
    Compute a datastructure in a state of static equilibrium using the force density method.
    """
    # compute static equilibrium
    eq_state = model(params, structure)

    # update equilibrium state in a copy of the datastructure
    return datastructure_updated(datastructure, eq_state, params)


def fdm(datastructure,
        sparse=True,
        is_load_local=False,
        tmax=1,
        eta=1e-6,
        itersolve_fn=None,
        implicit_diff=True,
        verbose=False):
    """
    Compute a datastructure in a state of static equilibrium using the force density method.
    """
    datastructure_validate(datastructure)

    model = model_from_sparsity(sparse, tmax, eta, is_load_local, itersolve_fn, implicit_diff, verbose)
    structure = structure_from_datastructure(datastructure, sparse)

    params = EquilibriumParametersState.from_datastructure(datastructure, dtype=DTYPE_JAX)

    return _fdm(model, params, structure, datastructure)


# ==========================================================================
# Constrained form-finding
# ==========================================================================

def constrained_fdm(datastructure,
                    optimizer,
                    loss,
                    parameters=None,
                    constraints=None,
                    maxiter=100,
                    tol=1e-6,
                    tmax=1,
                    eta=1e-6,
                    callback=None,
                    sparse=True,
                    is_load_local=False,
                    itersolve_fn=None,
                    implicit_diff=True,
                    verbose=False,
                    jit=True):
    """
    Generate a network in a constrained state of static equilibrium using the force density method.
    """
    datastructure_validate(datastructure)

    if constraints and sparse:
        print("Constraints are not supported yet for sparse inputs. Switching to dense.")
        sparse = False

    model = model_from_sparsity(sparse, tmax, eta, is_load_local, itersolve_fn, implicit_diff, verbose)
    structure = structure_from_datastructure(datastructure, sparse)

    opt_problem = optimizer.problem(model,
                                    structure,
                                    datastructure,
                                    loss,
                                    parameters,
                                    constraints,
                                    maxiter,
                                    tol,
                                    callback,
                                    jit)

    opt_params = optimizer.solve(opt_problem)

    params = optimizer.parameters_fdm(opt_params)

    return _fdm(model, params, structure, datastructure)


# ==========================================================================
# Helpers
# ==========================================================================

def model_from_sparsity(sparse, tmax, eta, is_load_local, itersolve_fn=None, implicit_diff=True, verbose=False):
    """
    Create an equilibrium model from a sparsity flag.
    """
    model = EquilibriumModel
    if sparse:
        model = EquilibriumModelSparse

    return model(tmax, eta, is_load_local, itersolve_fn, implicit_diff, verbose=verbose)


def structure_from_datastructure(datastructure, sparse):
    """
    Create a structure from a force density datastructure.
    """
    if isinstance(datastructure, FDNetwork):
        structure_factory = structure_from_network
    elif isinstance(datastructure, FDMesh):
        structure_factory = structure_from_mesh
    else:
        raise ValueError(f"Input datastructure {datastructure} is invalid")

    return structure_factory(datastructure, sparse)


def structure_from_network(network, sparse):
    """
    Create a structure from a network.
    """
    structure = EquilibriumStructure
    if sparse:
        structure = EquilibriumStructureSparse

    return structure.from_network(network)


def structure_from_mesh(mesh, sparse):
    """
    Create a structure from a mesh.
    """
    structure = EquilibriumMeshStructure
    if sparse:
        structure = EquilibriumMeshStructureSparse

    return structure.from_mesh(mesh)


def datastructure_validate(datastructure):
    """
    Check that the network is healthy.
    """
    assert datastructure.number_of_supports() > 0, "The FD datastructure has no supports"
    assert datastructure.number_of_edges() > 0, "The FD datastructure has no edges"
    assert np.all(np.abs(np.array(datastructure.edges_forcedensities())) > 0.0), "The FD datastructure has edges with zero force density"

    try:
        assert datastructure.number_of_nodes() > 0, "The FD datastructure has no nodes"
    except AttributeError:
        assert datastructure.number_of_vertices() > 0, "The FD datastructure has no nodes"


def datastructure_updated(datastructure, eq_state, params, use_loadsfromparams=False):
    """
    Return a copy of a datastructure whose attributes are updated with an equilibrium state.
    """
    datastructure = datastructure.copy()
    datastructure_update(datastructure, eq_state, params, use_loadsfromparams)

    return datastructure


def datastructure_update(datastructure, eq_state, params, use_loadsfromparams=False):
    """
    Update in-place the attributes of a datastructure with an equilibrium state.
    """
    # unpack equilibrium state
    xyz = eq_state.xyz.tolist()
    lengths = eq_state.lengths.tolist()
    residuals = eq_state.residuals.tolist()
    forces = eq_state.forces.tolist()

    forcedensities = params.q.tolist()

    if use_loadsfromparams:
        loads = params.loads.nodes.tolist()
    else:
        loads = eq_state.loads.tolist()

    # update edges
    datastructure_edges_update(datastructure,
                               (lengths, forces, forcedensities))

    # update nodes / vertices
    datastructure_nodes_update(datastructure,
                               (xyz, residuals, loads))


def datastructure_edges_update(datastructure, eqstate_edges):
    """
    Update the edge attributes of a datastructure.
    """
    lengths, forces, forcedensities = eqstate_edges

    for idx, edge in datastructure.index_uv().items():
        datastructure.edge_attribute(edge, name="length", value=lengths[idx].pop())
        datastructure.edge_attribute(edge, name="force", value=forces[idx].pop())
        datastructure.edge_attribute(edge, name="q", value=forcedensities[idx])


def datastructure_nodes_update(datastructure, eqstate_nodes):
    """
    Update the nodes or vertex attributes of a datastructure.
    """
    xyz, residuals, loads = eqstate_nodes

    if isinstance(datastructure, FDNetwork):
        nodevertex_updater = datastructure.node_attribute
    elif isinstance(datastructure, FDMesh):
        nodevertex_updater = datastructure.vertex_attribute
    else:
        raise ValueError(f"Input datastructure {datastructure} is invalid")

    for idx, key in datastructure.index_key().items():

        for name, value in zip("xyz", xyz[idx]):
            nodevertex_updater(key, name=name, value=value)

        for name, value in zip(["rx", "ry", "rz"], residuals[idx]):
            nodevertex_updater(key, name=name, value=value)

        for name, value in zip(["px", "py", "pz"], loads[idx]):
            nodevertex_updater(key, name=name, value=value)



import jax.numpy as jnp

from jax import vmap

from jax_fdm.goals import ScalarGoal

from jax_fdm.goals.mesh import MeshGoal

from jax_fdm.geometry import area_polygon


class MeshAreaGoal(ScalarGoal, MeshGoal):
    """
    Maximize the negative area of a mesh.
    """
    def __init__(self, target=0.0, weight=1.0):
        super().__init__(key=-1, target=target, weight=weight)
        self.faces = None

    def init(self, model, structure):
        """
        Initialize the goal with information of a model and a structure.
        """
        super().init(model, structure)
        self.faces = structure.faces_indexed

    def prediction(self, eq_state, index):
        """
        The current load path of the network.
        """
        def face_xyz(face, xyz):
            """
            Get this face XYZ coordinates from XYZ vertices array.
            """
            face = jnp.ravel(face)

            xyz_face = xyz[face, :]
            xyz_repl = xyz_face[0, :]

            # NOTE: Replace -1 with first entry to avoid nans in gradient computation
            # This was a pesky bug, since using nans as replacement did not cause
            # issues with the forward computation of normals, but it does for
            # the backward pass.
            xyz_face = vmap(jnp.where, in_axes=(0, 0, None))(face >= 0, xyz_face, xyz_repl)

            return xyz_face

        def face_area(face, xyz):
            fxyz = face_xyz(face, xyz)
            return area_polygon(fxyz)

        faces_area = vmap(face_area, in_axes=(0, None))
        areas = faces_area(self.faces, eq_state.xyz)

        area = jnp.sum(areas) * -1.0

        return jnp.atleast_1d(area)


class MeshFacesAreaEqualizeGoal(ScalarGoal, MeshGoal):
    """
    Maximize the negative area of a mesh.
    """
    def __init__(self, target=0.0, weight=1.0):
        super().__init__(key=-1, target=target, weight=weight)
        self.faces = None

    def init(self, model, structure):
        """
        Initialize the goal with information of a model and a structure.
        """
        super().init(model, structure)
        self.faces = structure.faces_indexed

    def prediction(self, eq_state, index):
        """
        The current load path of the network.
        """
        def face_xyz(face, xyz):
            """
            Get this face XYZ coordinates from XYZ vertices array.
            """
            face = jnp.ravel(face)

            xyz_face = xyz[face, :]
            xyz_repl = xyz_face[0, :]
            xyz_face = vmap(jnp.where, in_axes=(0, 0, None))(face >= 0, xyz_face, xyz_repl)

            return xyz_face

        def face_area(face, xyz):
            fxyz = face_xyz(face, xyz)
            return area_polygon(fxyz)

        faces_area = vmap(face_area, in_axes=(0, None))
        areas = faces_area(self.faces, eq_state.xyz)

        return jnp.atleast_1d(jnp.var(areas) / jnp.mean(areas))



import inspect

from collections import defaultdict


class Collection:
    """
    A collection of goals of the same type to speed up optimization.
    """
    def __new__(cls, collectibles):
        """
        Gather a collection of objects into a single vectorized object.
        """
        # check homogenity
        ctypes = [type(c) for c in collectibles]
        if len(set(ctypes)) != 1:
            raise TypeError("The input collectibles are not of the same type!")

        # extract class
        cls = ctypes.pop()

        # class signature
        sig = inspect.signature(cls)

        # collect init signature values
        ckwargs = defaultdict(list)
        for key in sig.parameters.keys():
            for collectible in collectibles:
                attr = getattr(collectible, key)
                ckwargs[key].append(attr)

        collection = cls(**ckwargs)
        collection._iscollection = True

        return collection



from .node import *  # noqa F403
from .coordinates import *  # noqa F403
from .normal import *  # noqa F403
from .tangent import *  # noqa F403
from .curvature import *  # noqa F403


__all__ = [name for name in dir() if not name.startswith('_')]



from .edge import *  # noqa F403
from .length import *  # noqa F403
from .force import *  # noqa F403
from .loadpath import *  # noqa F403
from .direction import *  # noqa F403
from .angle import *  # noqa F403


__all__ = [name for name in dir() if not name.startswith('_')]



"""
A force density mesh.
"""
from math import fabs
from statistics import stdev

from compas.datastructures import Datastructure


class FDDatastructure(Datastructure):
    """
    A force density datastructure.
    """
    # ----------------------------------------------------------------------
    # Edges
    # ----------------------------------------------------------------------

    def edge_load(self, key, load=None):
        """
        Gets or sets a load on an edge.
        """
        return self.edge_attributes(key, names=("px", "py", "pz"), values=load)

    def edge_forcedensity(self, key, q=None):
        """
        Gets or sets the force density on a single edge.
        """
        return self.edge_attribute(key, name="q", value=q)

    def edge_force(self, key):
        """
        Gets the forces at a single edge the network.
        """
        return self.edge_attribute(key, name="force")

    def edge_loadpath(self, key):
        """
        Gets the load path at a single edge the network.
        """
        force = self.edge_attribute(key, name="force")
        length = self.edge_attribute(key, name="length")
        return fabs(force * length)

    def edges_forcedensities(self, q=None, keys=None):
        """
        Gets or sets the force densities on a list of edges.
        """
        return self.edges_attribute(name="q", value=q, keys=keys)

    def edges_forces(self, keys=None):
        """
        Gets the forces on the edges of the network.
        """
        return self.edges_attribute(keys=keys, name="force")

    def edges_lengths(self, keys=None):
        """
        Gets the lengths on the edges of the network.
        """
        return self.edges_attribute(keys=keys, name="length")

    def edges_loads(self, load=None, keys=None):
        """
        Gets or sets a load to the edges of the datastructure.
        """
        return self.edges_attributes(names=("px", "py", "pz"), values=load, keys=keys)

    def edges_loadpaths(self, keys=None):
        """
        Gets the load path on the edges of the network.
        """
        keys = keys or self.edges()
        for key in keys:
            yield self.edge_loadpath(key)

    # ----------------------------------------------------------------------
    # Datastructure properties
    # ----------------------------------------------------------------------

    def loadpath(self):
        """
        Gets the total load path of the network.
        """
        return sum(list(self.edges_loadpaths()))

    def print_stats(self, other_stats=None):
        """
        Print information aboud the equilibrium state of the network.
        """
        stats = {"FDs": self.edges_forcedensities(),
                 "Forces": self.edges_forces(),
                 "Lengths": self.edges_lengths()}

        other_stats = other_stats or {}
        stats.update(other_stats)

        print(f"\n***{self.__class__.__name__} stats***")
        print(f"Load path: {round(self.loadpath(), 3)}")

        for name, vals in stats.items():

            minv = round(min(vals), 3)
            maxv = round(max(vals), 3)
            meanv = round(sum(vals) / len(vals), 3)
            stdv = round(stdev(vals), 3)
            name = "{:<18}".format(name)

            print(f"{name}\tMin: {minv}\tMax: {maxv}\tMean: {meanv}\tStDev: {stdv}")



import numpy as np

import jax.numpy as jnp

from jax import vmap

from jax_fdm.geometry import angle_vectors
from jax_fdm.geometry import normalize_vector

from jax_fdm.constraints.node import NodeConstraint


class NodeNormalAngleConstraint(NodeConstraint):
    """
    Constraints the angle formed by the node normal and a vector between a lower and an upper bound.
    """
    def __init__(self, key, vector, bound_low=None, bound_up=None):
        super().__init__(key=key, bound_low=bound_low, bound_up=bound_up)
        self._vector = None
        self.vector = vector

        self.index_faces = None
        self.mask_faces = None

    @property
    def vector(self):
        """
        The vector to take the angle with.
        """
        return self._vector

    @vector.setter
    def vector(self, vector):
        self._vector = jnp.reshape(jnp.asarray(vector), (-1, 3))

    def vectors(self):
        """
        Create a matrix of vectors.
        """
        matrix = np.zeros((max(self.index) + 1, 3))
        for vec, idx in zip(self.vector, self.index):
            matrix[idx, :] = vec
        return matrix

    def init(self, model, structure):
        """
        Initialize the constraint with information from an equilibrium model.
        """
        super().init(model, structure)
        self.vector = self.vectors()
        self.index_faces, self.mask_faces = self.faces_indices(model, structure)

    def faces_indices(self, model, structure):
        """
        Get the node indices of the faces connected to a node.

        TODO: Please refactor me! I am illegible :(
        """
        index_max = max(self.index) + 1

        num_nodes_max = -1
        num_faces_max = -1
        for idx in self.index:
            num_faces = len(self.node_faces(model, structure, idx))
            if num_faces > num_faces_max:
                num_faces_max = num_faces
            num_nodes = len(max(self.node_faces_indices(model, structure, idx), key=lambda x: len(x)))
            if num_nodes > num_nodes_max:
                num_nodes_max = num_nodes

        index_faces = np.zeros((index_max, num_faces_max, num_nodes_max + 1))
        mask_faces = np.zeros((index_max, num_faces_max, num_nodes_max + 1))

        for idx in self.index:
            nfn = self.node_faces_indices(model, structure, idx)
            for i, face_nodes in enumerate(nfn):
                index_faces[idx, i, :len(face_nodes)] = face_nodes
                mask_faces[idx, i, :len(face_nodes)] = [1] * len(face_nodes)

                index_faces[idx, i, -1] = face_nodes[-1]
                mask_faces[idx, i, -1] = 1

        index_faces = jnp.asarray(index_faces, dtype=jnp.int64)
        mask_faces[mask_faces == 0] = jnp.nan
        mask_faces = jnp.asarray(mask_faces, dtype=jnp.float64)

        return index_faces, mask_faces

    @staticmethod
    def node_faces(model, structure, index):
        """
        Return an iterable with the indices of the faces connected to a node.
        """
        connectivity = structure.connectivity_faces
        return np.flatnonzero(connectivity[:, index])

    def node_faces_indices(self, model, structure, index):
        """
        Return an iterable with the indices of the nodes of the faces connected to a node.
        """
        fidx = self.node_faces(model, structure, index)
        return [structure.face_node_index[idx] for idx in fidx]

    @staticmethod
    def nan_normal_polygon(polygon):
        """
        Compute the normal of a polygon that contains nan entries.
        """
        centroid = jnp.nanmean(polygon, axis=0)
        op = polygon - centroid
        op_shifted = jnp.roll(op, 1, axis=0)
        ns = 0.5 * jnp.cross(op_shifted, op)

        return jnp.nansum(ns, axis=0)

    def face_normal(self, eqstate, face, mask):
        """
        Calculate the normal of a face.
        """
        polygon = eqstate.xyz[face, :]
        mask = jnp.reshape(mask, (-1, 1))
        polygon = mask * polygon

        return self.nan_normal_polygon(polygon)

    def node_normal(self, eqstate, index):
        """
        Get the unitized normal vector at a node.
        """
        faces = self.index_faces[index]
        masks = self.mask_faces[index]
        normals = vmap(self.face_normal, in_axes=(None, 0, 0))(eqstate, faces, masks)

        return normalize_vector(jnp.nanmean(normals, axis=0))

    def constraint(self, eqstate, index):
        """
        Returns the normal at a node based on the xyz coordinates of its one-hop neighborhood.
        """
        normal = self.node_normal(eqstate, index)

        return angle_vectors(normal, self.vector[index, :])



"""
NOTE: Sparse solver does not support forward mode auto-differentiation yet.
"""
import jax
import jax.numpy as jnp

from scipy.sparse import csc_matrix
from scipy.sparse.linalg import spsolve as spsolve_scipy

from jax.experimental.sparse import CSC
from jax.experimental.sparse.linalg import spsolve as spsolve_jax


# ==========================================================================
# Sparse linear solver on GPU
# ==========================================================================

def spsolve_gpu_ravel(A, b):
    """
    A wrapper around cuda sparse linear solver that is GPU friendly.

    Notes
    -----
    We ravel b into a single vector because the CUDA sparse solver backend
    cannot take b as a matrix.
    Accordingly, we must construct a sparse block diagonal matrix that
    repeats matrix A three times (one time per spatial coordinate XYZ).

    This "raveled" linear solver could be faster than the `spsolve_gpu_stack`
    alternative, because this functions solves the block linear system
    all at once instead of 3 times in sequence.

    This limitation with CUDA might make a GPU sparse solve more expensive
    than a CPU sparse solve. So use this method with a pinch of salt.
    """
    # NOTE: we can pass CSC indices directly to the JAX solver because we can!
    # Just kidding. This is because the matrix A is symmetric :)
    A = blockdiag_matrix_sparse(A, 3)
    b = jnp.ravel(b, "F")
    X = spsolve_jax(A.data, A.indices, A.indptr, b)

    return jnp.reshape(X, (3, -1)).T


def spsolve_gpu_stack(A, b):
    """
    A wrapper around cuda sparse linear solver that is GPU friendly.

    Notes
    -----
    We must split b into three vectors (one per spatial coordinate XYZ),
    and then solve a sparse system 3 times because the CUDA sparse solver
    in the backend cannot take b as a matrix, but only as a vector.

    The sparse solve in a GPU could thus be 3x more expensive than if it could
    solve for the b matrix all at once?

    This limitation with CUDA might make a GPU sparse solve more expensive
    than a CPU sparse solve. So use this method with a pinch of salt.
    """
    # NOTE: JAX requires a csr matrix as input,
    # but we can pass csc indices directly because we can!
    # Just kidding. This is because the matrix A is symmetric :)

    # TODO: Ravel and unravel this!
    x = spsolve_jax(A.data, A.indices, A.indptr, b[:, 0])
    y = spsolve_jax(A.data, A.indices, A.indptr, b[:, 1])
    z = spsolve_jax(A.data, A.indices, A.indptr, b[:, 2])

    return jnp.stack((x, y, z), axis=1)


# Set ravel solver as the GPU sparse solver
spsolve_gpu = spsolve_gpu_ravel


# ==========================================================================
# Sparse linear solver on CPU
# ==========================================================================

def spsolve_cpu(A, b):
    """
    A wrapper around scipy sparse linear solver that acts as a JAX pure callback.
    """
    def callback(data, indices, indptr, _b):
        _A = csc_matrix((data, indices, indptr))
        return spsolve_scipy(_A, _b)

    xk = jax.pure_callback(callback,  # callback function
                           b,  # return type is b
                           A.data,  # callback function arguments from here on
                           A.indices,
                           A.indptr,
                           b)
    return xk


# ==========================================================================
# Register sparse linear solvers
# ==========================================================================

def register_sparse_solver(solvers):
    """
    Register the sparse solver used by the FDM model based on JAX default backend.
    """
    backend = jax.default_backend()
    sparse_solver = solvers.get(backend)

    if not sparse_solver:
        raise ValueError(f"Default backend {backend} does not support a sparse solver")

    return sparse_solver


solvers = {"cpu": spsolve_cpu,
           "gpu": spsolve_gpu}

spsolve = register_sparse_solver(solvers)


# ==========================================================================
# Define sparse linear solver
# ==========================================================================

@jax.custom_vjp
def sparse_solve(A, b):
    """
    The sparse linear solver.
    """
    return spsolve(A, b)


# ==========================================================================
# Forward and backward passes
# ==========================================================================

def sparse_solve_fwd(A, b):
    """
    Forward pass of the sparse linear solver.

    Call the linear solve and save parameters and solution for the backward pass.
    """
    xk = sparse_solve(A, b)

    return xk, (xk, A, b)


def sparse_solve_bwd(res, g):
    """
    Backward pass of the sparse linear solver.
    """
    xk, A, b = res

    # Solve adjoint system
    # A.T @ xk_bar = -g
    lam = sparse_solve(A, g)

    # the implicit constraint function for implicit differentiation
    def residual_fn(params):
        A, b = params
        return b - A @ xk

    params = (A, b)

    # Call vjp of residual_fn to compute gradient wrt params
    params_bar = jax.vjp(residual_fn, params)[1](lam)[0]

    return params_bar


# ==========================================================================
# Sparse matrix helpers
# ==========================================================================

def blockdiag_matrix_sparse(A, num=2, format=CSC):
    """
    Build a block diagonal sparse matrix in the input format by repeating
    a square sparse matrix a prescribed number of times.
    """
    indptr = []
    indices = []
    data = []

    assert num > 1, "Block diagonal matrix must have at least 2 blocks."

    nrows, ncols = A.shape
    assert nrows == ncols, "Supported sparse matrices must be square."

    c_idx = 0
    r_idx = 0
    for i in range(num):
        _indptr = A.indptr + c_idx
        if i > 0:
            _indptr = _indptr[1:]

        indptr.append(_indptr)
        indices.append(A.indices + r_idx)
        data.append(A.data)

        c_idx += jnp.max(A.indptr)
        r_idx += nrows

    indptr = jnp.concatenate(indptr)
    indices = jnp.concatenate(indices)
    data = jnp.concatenate(data)

    return format((data, indices, indptr),
                  shape=(ncols*num, nrows*num))


# ==========================================================================
# Register forward and backward passes to JAX
# ==========================================================================

sparse_solve.defvjp(sparse_solve_fwd, sparse_solve_bwd)



from .network_artist import *  # noqa F403
from .viewer import *  # noqa F403
from .register import register_artists


register_artists()

__all__ = [name for name in dir() if not name.startswith('_')]



import numpy as np

from jax_fdm.goals import GoalState


# ==========================================================================
# Goal manager
# ==========================================================================

def goals_reindex(goals, model):
    """
    Compute the index of a goal based on its node or edge key.
    """
    for goal in goals:
        index = goal.index_from_model(model, goal.key)
        goal.index = index


def goals_state(goals, eqstate, model):
    """
    Collate goals attributes into vectors.
    """
    predictions = []
    targets = []
    weights = []

    for goal in goals:
        gstate = goal(eqstate, model)
        predictions.append(gstate.prediction)
        targets.append(gstate.target)
        weights.append(gstate.weight)

    predictions = np.concatenate(predictions, axis=0)
    targets = np.concatenate(targets, axis=0)
    weights = np.concatenate(weights, axis=0)

    return GoalState(prediction=predictions, target=targets, weight=weights)



import jax.numpy as jnp

from jax_fdm.goals import ScalarGoal

from jax_fdm.goals.network import NetworkGoal


class NetworkXYZLaplacianGoal(ScalarGoal, NetworkGoal):
    """
    Minimize the Laplacian energy of the XYZ coordinates of a network.

    Notes
    -----
    This goal can be handy to "smoothen" the looks of a network.

    The Laplacian energy of some feature vector x is defined as
    E = x^T @ L @ x, where the unnormalized Laplacian L is related to
    the connectivity matrix of a network C such that L = C^T @ C.

    Given the above relationship and that we want to calculate the
    energy of the XYZ coordinates of a network in equilibrium
    (the feature vector x of interest are the XYZ coordinates),
    then the energy expression becomes: x^T @ (C^T @ C) @ X.

    We know that U = C @ X is the matrix with coordinate difference
    vectors that we obtain from the equilibrium model.

    Therefore, we can re-rexpress the energy E in terms of U, and
    use that definition to calculate the energy for this goal
    function: E = U^T @ U. Interestingly, this expression is
    equivalent to minimizing the squared sum of the edge lengths l
    of the network, E = l^2.

    In practice, we are interested in a single scalar value of
    this energy to minimize with optimization. As a result, we
    ravel matrix U into a single vector, and calculate the dot
    product of U in the raveled state.
    """
    @staticmethod
    def prediction(eq_state, index):
        """
        The current load path of the network.
        """
        vectors = eq_state.vectors
        laplacian = jnp.dot(jnp.ravel(vectors),
                            jnp.reshape(vectors, (-1, 1)))

        return jnp.atleast_1d(laplacian)



from math import fabs

from compas.geometry import add_vectors
from compas.geometry import length_vector
from compas.geometry import scale_vector
from compas.geometry import Point
from compas.geometry import Vector

from compas_plotters.artists import NetworkArtist

from jax_fdm.visualization.artists import FDNetworkArtist


class FDNetworkPlotterArtist(FDNetworkArtist, NetworkArtist):
    """
    An artist that knows how to draw a force density network in a plotter.
    """
    def draw_nodes(self):
        """
        Draw the nodes of the network.
        """
        return NetworkArtist.draw_nodes(self)

    def draw_edges(self):
        """
        Draw the edges of the network.
        """
        return NetworkArtist.draw_edges(self)

    def draw_reaction(self, node, scale, color, *args, **kwargs):
        """
        Draw a reaction vector at a node.
        """
        network = self.network

        vector = network.node_reaction(node)
        start = network.node_coordinates(node)

        if length_vector(vector) < self.reaction_tol:
            return

        # shift starting point if max force of connected edges is compressive
        forces = [network.edge_force(e) for e in network.connected_edges(node)]
        max_force = max(forces, key=lambda f: fabs(f))
        if max_force < 0.0:
            start = add_vectors(start, scale_vector(vector, scale))

        # reverse vector to display direction of reaction forces
        reaction = self.draw_vector(scale_vector(vector, -1.0), start, scale)

        return self.plotter.add(reaction, point=Point(*start), color=color)

    def draw_load(self, node, scale, color, *args, **kwargs):
        """
        Draw a load vector at a node.
        """
        vector = self.network.node_load(node)

        if length_vector(vector) < self.load_tol:
            return

        start = self.network.node_coordinates(node)
        load = self.draw_vector(vector, start, scale)
        start = add_vectors(start, scale_vector(load, -1.))

        return self.plotter.add(load, point=Point(*start), color=color)

    @staticmethod
    def draw_vector(vector, start, scale):
        """
        Draw a vector as an arrow.
        """
        vector_scaled = scale_vector(vector, scale)
        end = add_vectors(start, vector_scaled)

        return FDVector.from_start_end(start, end)


class FDVector(Vector):
    """
    A wrapper around a compas.geometry.Vector artist.
    The goal if this artist to override how vectors are drawn in a plotter.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



"""
********************************************************************************
jax_fdm
********************************************************************************

.. currentmodule:: jax_fdm


.. toctree::
    :maxdepth: 1


"""

from __future__ import print_function

import os

import numpy as np
import jax.numpy as jnp


__author__ = ["Rafael Pastrana"]
__copyright__ = "Rafael Pastrana"
__license__ = "MIT License"
__email__ = "arpastrana@princeton.edu"
__version__ = "0.8.5"


HERE = os.path.dirname(__file__)

HOME = os.path.abspath(os.path.join(HERE, "../../"))
DATA = os.path.abspath(os.path.join(HOME, "data"))
DOCS = os.path.abspath(os.path.join(HOME, "docs"))
TEMP = os.path.abspath(os.path.join(HOME, "temp"))

__all__ = ["HOME", "DATA", "DOCS", "TEMP"]

# config.py
# define floating point precision
DTYPE_NP = np.float64
DTYPE_JAX = jnp.float64

DTYPE_INT_NP = jnp.int32
DTYPE_INT_JAX = jnp.int32

# this only works on startup!
if DTYPE_JAX == jnp.float64 or DTYPE_NP == np.float64:
    import jax
    jax.config.update("jax_enable_x64", True)



from compas.artists import Artist

from jax_fdm.datastructures import FDNetwork

from jax_fdm.visualization.notebooks import FDNetworkNotebookArtist


__all__ = ["register_artists"]


def register_artists():
    """
    Register objects to the artist factory.
    """
    Artist.register(FDNetwork, FDNetworkNotebookArtist, context="Notebook")



import jax.numpy as jnp

from jax.lax import cond


WORLD_X = jnp.array([1.0, 0.0, 0.0])
WORLD_Y = jnp.array([0.0, 1.0, 0.0])
WORLD_Z = jnp.array([0.0, 0.0, 1.0])
WORLD_XYZ = jnp.eye(3)


def angle_vectors(u, v, deg=False):
    """
    Compute the smallest angle in degrees between two vectors.
    """
    L = length_vector(u) * length_vector(v)
    cosim = jnp.maximum(jnp.minimum((u @ v) / L, 1.0), -1.0)
    angle = jnp.arccos(cosim)

    if deg:
        return jnp.degrees(angle)
    return angle


def length_vector(u):
    """
    Calculate the length of a vector along its columns.
    """
    length = jnp.linalg.norm(u, axis=-1, keepdims=True)

    return length


def normalize_vector(u, safe_nan=True):
    """
    Scale a vector such that it has a unit length.
    """
    if safe_nan:
        u = jnp.nan_to_num(u)
        is_zero_vector = jnp.allclose(u, 0.0)
        u = jnp.where(is_zero_vector, jnp.ones_like(u), u)
        length = jnp.where(is_zero_vector, 1.0, length_vector(u))
    else:
        length = length_vector(u)

    return u / length


def vector_unitized(u):
    """
    Scale a vector such that it has a unit length.
    """
    return u / length_vector(u)


def subtract_vectors(u, v):
    """
    Subtract two vectors.
    """
    return u - v


def line_vector(line, normalized=True):
    """
    Calculate the (normalized) vector formed by the difference of the line end points.
    """
    vector = subtract_vectors(line[1], line[0])
    if normalized:
        return vector_unitized(vector)

    return vector


def vector_projection(u, v):
    """
    Calculates the orthogonal projection of u onto v.
    """
    l2 = jnp.sum(v ** 2)
    x = (u @ jnp.transpose(v)) / l2

    return v * x


def closest_point_on_plane(point, plane):
    """
    Computes the closest location on a plane to a supplied point.
    """
    origin, normal = plane
    normal = normalize_vector(normal)
    d = normal @ origin
    e = normal @ point - d
    k = e / jnp.sum(jnp.square(normal))

    return point - normal * k


def closest_point_on_line(point, line):
    """
    Computes the closest location on a line to a supplied point.
    """
    a, b = line
    ab = b - a
    ap = point - a
    c = vector_projection(ap, ab)

    return a + c


def closest_point_on_segment(point, segment):
    """
    Calculate the closest location on a segment to an input point.
    """
    a, b = segment
    p = closest_point_on_line(point, segment)

    # calculate distances to the three possible points
    d = distance_point_point_sqrd(a, b)
    d1 = distance_point_point_sqrd(a, p)
    d2 = distance_point_point_sqrd(b, p)

    # define callables to be compatible with the signature of lax.cond
    def start_point():
        return a

    def end_point():
        return b

    def middle_point():
        return p

    def point_is_outside():
        return cond(d1 < d2, start_point, end_point)

    # return closest amont the three points
    return cond(jnp.logical_or(d1 > d, d2 > d), point_is_outside, middle_point)


def distance_point_point_sqrd(u, v):
    """
    Calculate the square of the distance between two points.
    """
    vector = subtract_vectors(u, v)

    return jnp.sum(jnp.square(vector))


def normal_polygon_2(polygon, unitized=False):
    """
    Computes the unit-length normal of a polygon.

    A polygon that is defined as a sequence of unique points.
    A polygon must have at least three points.
    """
    centroid = jnp.mean(polygon, axis=0)
    op = polygon - centroid
    op_shifted = jnp.roll(op, 1, axis=0)
    ns = 0.5 * jnp.cross(op_shifted, op)
    n = jnp.sum(ns, axis=0)

    if unitized:
        return normalize_vector(n)
    return n


def normal_polygon(polygon, unitized=True):
    """
    Computes the normal of a closed polygon.

    A polygon that is defined as a sequence of unique points and must be
    defined by at least 3 points.

    This function ignores nan values if any in the input polygon.
    """
    assert len(polygon) >= 3, "A polygon must be defined by at least 3 points"

    polygon_shifted = jnp.roll(polygon, 1, axis=0)
    ns = 0.5 * jnp.cross(polygon_shifted, polygon)
    normal = jnp.nansum(ns, axis=0)

    if unitized:
        normal = normalize_vector(normal)

    return normal


def area_polygon(polygon):
    """
    Computes the area of a polygon.

    A polygon that is defined as a sequence of unique points and must be
    defined by at least 3 points.

    This function ignores nan values if any in the input polygon.
    """
    return length_vector(normal_polygon(polygon, unitized=False))


def normal_triangle(triangle, unitize=False):
    """
    Computes the normal vector of a triangle.

    A triangle is defined as a set of exactly three points.
    """
    assert len(triangle) == 3, "A triangle is defined by exactly 3 points"

    line_a, line_b = triangle[:-1, :] - triangle[1:, :]
    normal = jnp.cross(line_a, line_b)

    if unitize:
        return normalize_vector(normal)

    return normal


def area_triangle(triangle):
    """
    Calculate the area of a triangle.

    A triangle is defined as a set of exactly three points.
    """
    return 0.5 * length_vector(normal_triangle(triangle))


def _curvature_point_polygon(point, polygon):
    raise NotImplementedError


def curvature_point_polygon(point, polygon):
    """
    Compute the discrete curvature at a point based on a polygon surrounding it.
    The discrete curvature of a node equals 2 * pi - sum(alphas).

    TODO: divide return value by tributary area of point.

    Notes
    -----
    Alphas is the list of angles between each pair of successive edges as
    the outward vectors from the node. Polygon is a numpy array (# points, 3).
    """
    op = polygon - point
    norm_op = jnp.reshape(jnp.linalg.norm(op, axis=1), (-1, 1))
    op = op / norm_op
    op_off = jnp.roll(op, 1, axis=0)
    dot = jnp.sum(op_off * op, axis=1)
    dot = jnp.maximum(jnp.minimum(dot, jnp.full(dot.shape, 1)), jnp.full(dot.shape, -1))
    angles = jnp.arccos(dot)

    return 2 * jnp.pi - jnp.sum(angles)


def line_lcs(line):
    """
    Returns the local coordinate system (LCS) of a line.

    The LCS is a 3D orthonormal basis formed by unit vectors U, V, and W.
    The orthonormal basis is constructed using the following convention:

    - The U axis is the unit vector of the coordinate difference of the line end points.
    - The V axis is the cross product of U and the global Z axis. If the V axis is
      parallel to Z, then V becomes the cross product of U and the global Y axis.
    - The W axis is the cross product of U and V.
    """
    u = line_vector(line)

    threshold = jnp.allclose(jnp.abs(WORLD_Z @ u), 1.0)
    vperp = jnp.where(threshold, WORLD_Y, WORLD_Z)
    v = jnp.cross(vperp, u)

    w = jnp.cross(u, v)
    w = jnp.where(w @ vperp < 0.0, -w, w)

    return jnp.vstack((u, v, w))


def polygon_lcs(polygon):
    """
    Returns the local coordinate system (LCS) of a polygon.

    The LCS is a 3D orthonormal basis formed by unit vectors U, V, and W.
    The orthonormal basis is constructed using the following convention:

    - The W axis is the polygon normal.
    - The U axis is the unit vector of the coordinate difference of the line end points.
    - The V axis is the cross product of U and the global Z axis. If the V axis is
        parallel to Z, then V becomes the cross product of U and the global Y axis.
    """
    w = normal_polygon(polygon, True)

    threshold = jnp.allclose(jnp.abs(WORLD_X @ w), 1.0)
    vperp = jnp.where(threshold, WORLD_Y, WORLD_X)
    v = jnp.cross(w, vperp)

    u = jnp.cross(w, v)
    u = jnp.where(u @ vperp < 0.0, -u, u)

    return jnp.vstack((u, v, w))


if __name__ == "__main__":

    from compas.geometry import Polygon
    from compas.geometry import Rotation
    from compas.geometry import Frame
    from compas.geometry import Vector

    from jax import jit
    from jax import grad

    from jax.config import config
    config.update("jax_debug_nans", True)

    # Test vector transformation from XYZ to polygon LCS

    # Test polygon LCS
    load = Vector(0.0, 0.0, 1.0)
    load_scale = 2.0
    polygon = Polygon.from_sides_and_radius_xy(4, 1.0)

    polygon_lcs = jit(polygon_lcs)
    lcs = polygon_lcs(jnp.array(polygon.points))
    assert jnp.allclose(lcs, WORLD_XYZ), f"lcs:\n{lcs}\nlcs_target:\n{WORLD_XYZ}"

    load_lcs = jnp.array(load * load_scale) @ lcs
    load_target = WORLD_Z * load_scale
    assert jnp.allclose(load_lcs, load_target), f"lcs:\n{load_lcs}\nlcs_target:\n{load_target}"
    load_lcs = -jnp.array(load * load_scale) @ lcs
    load_target = load_target * -1.0
    assert jnp.allclose(load_lcs, load_target), f"lcs:\n{load_lcs}\nlcs_target:\n{load_target}"
    print("Passed world XYZ test\n")

    R = Rotation.from_frame_to_frame(Frame.worldXY(), Frame.worldZX())
    polygon_transformed = polygon.transformed(R)
    lcs = polygon_lcs(jnp.array(polygon_transformed.points[::-1]))
    lcs_target = jnp.vstack((WORLD_X, WORLD_Z, -WORLD_Y))
    assert jnp.allclose(lcs, lcs_target), f"lcs:\n{lcs}\nlcs_target:\n{lcs_target}"

    load_lcs = jnp.array(load * load_scale) @ lcs
    load_target = -WORLD_Y * load_scale
    assert jnp.allclose(load_lcs, load_target), f"lcs:\n{load_lcs}\nlcs_target:\n{load_target}"
    load_lcs = -jnp.array(load * load_scale) @ lcs
    load_target = load_target * -1.0
    assert jnp.allclose(load_lcs, load_target), f"lcs:\n{load_lcs}\nlcs_target:\n{load_target}"

    print("Passed world XY to ZX test\n")

    R = Rotation.from_frame_to_frame(Frame.worldXY(), Frame.worldYZ())
    polygon_transformed = polygon.transformed(R)
    lcs = polygon_lcs(jnp.array(polygon_transformed.points))
    lcs_target = jnp.vstack((WORLD_Y, WORLD_Z, WORLD_X))
    assert jnp.allclose(lcs, lcs_target), f"lcs:\n{lcs}\nlcs_target:\n{lcs_target}"

    load_lcs = jnp.array(load * load_scale) @ lcs
    load_target = WORLD_X * load_scale
    assert jnp.allclose(load_lcs, load_target), f"lcs:\n{load_lcs}\nlcs_target:\n{load_target}"
    load_lcs = -jnp.array(load * load_scale) @ lcs
    load_target = load_target * -1.0
    assert jnp.allclose(load_lcs, load_target), f"lcs:\n{load_lcs}\nlcs_target:\n{load_target}"

    print("Passed world XY to YZ test\n")

    # Test line LCS
    line_lcs = jit(line_lcs)
    line = [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]]
    line = jnp.array(line)
    lcs = line_lcs(line)

    lcs_target = WORLD_XYZ
    assert jnp.allclose(lcs, lcs_target), f"lcs:\n{lcs}\nlcs_target:\n{lcs_target}"

    line = [[0.0, 0.0, 0.0], [0.0, 1.0, 0.0]]
    line = jnp.array(line)
    lcs = line_lcs(line)

    lcs_target = jnp.array([[0.0, 1.0, 0.0], [-1.0, 0.0, 0.0], [0.0, 0.0, 1.0]])
    assert jnp.allclose(lcs, lcs_target), f"lcs:\n{lcs}\nlcs_target:\n{lcs_target}"

    line = [[0.0, 0.0, 0.0], [0.0, 0.0, 1.0]]
    line = jnp.array(line)
    lcs = line_lcs(line)

    lcs_target = jnp.array([[0.0, 0.0, 1.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0]])
    assert jnp.allclose(lcs, lcs_target), f"lcs:\n{lcs}\nlcs_target:\n{lcs_target}"

    # Test area polygon
    import numpy as np
    from jax import vmap, jit
    from math import pi
    from compas.geometry import Polygon
    from compas.geometry import Rotation
    from compas.geometry import area_polygon as compas_area_polygon
    from compas.geometry import normal_polygon as compas_normal_polygon

    radius = 2.0
    # num_angles = 100
    num_angles = 4

    polygons = []

    area_polygon = jit(area_polygon)
    area_polygon(np.ones((5, 3)))

    grad_area_normal_polygon = grad(lambda x: jnp.sum(jnp.square(normal_polygon(x) - 1.0)))
    grad_area_normal_polygon = jit(grad_area_normal_polygon)

    for i in range(4, 5):
        polygon = Polygon.from_sides_and_radius_xy(i, radius)

        for j in range(1, num_angles):

            polygon_old = polygon

            angle = (j * (0.5 * pi / num_angles))
            R1 = Rotation.from_axis_and_angle(axis=[1.0, 0.0, 0.0], angle=angle)
            polygon = polygon.transformed(R1)
            R2 = Rotation.from_axis_and_angle(axis=[0.0, 1.0, 0.0], angle=angle)
            polygon = polygon.transformed(R2)
            polygons.append(polygon)

            assert not np.allclose(polygon, polygon_old)

            area_compas = compas_area_polygon(polygon.points)
            area = area_polygon(np.array(polygon.points))  # .item()

            assert jnp.allclose(area_compas, area), f"Not equal: JAX: {area:.2f} vs. COMPAS: {area_compas:.2f}"

            normal_compas = compas_normal_polygon(polygon.points)
            normal = normal_polygon(np.array(polygon.points))  # .item()
            assert jnp.allclose(np.array(normal_compas), normal), f"Not equal: JAX: {normal:.2f} vs. COMPAS: {normal_compas:.2f}"

            points_nan = np.reshape(np.array([np.nan] * 6), (-1, 3))
            polygon_nan = np.vstack((np.array(polygon.points), points_nan, points_nan))
            normal_nan = normal_polygon(polygon_nan)

            assert jnp.allclose(np.array(normal_compas), normal_nan), f"Not equal: JAX: {normal_nan:.2f} vs. COMPAS: {normal_compas:.2f}"

            points_padlast = np.array(polygon.points)[0, :]
            polygon_padlast = np.vstack((np.array(polygon.points), points_padlast))
            normal_padlast = normal_polygon(polygon_padlast)

            assert jnp.allclose(np.array(normal_compas), normal_padlast), f"Not equal: JAX: {normal_nan:.2f} vs. COMPAS: {normal_padlast:.2f}"

            gnormal_nan = grad_area_normal_polygon(polygon_nan)
            assert jnp.sum(jnp.isnan(gnormal_nan)) == 0, "Gradient of nan'd polygon has nans!"

            gnormal_padlast = grad_area_normal_polygon(polygon_padlast)

            print(polygon_nan)
            print(gnormal_nan)
            print()
            print(polygon_padlast)
            print(gnormal_padlast)
            print()

            assert jnp.sum(jnp.isnan(gnormal_padlast)) == 0, "Gradient of padded polygon has nans!"

    areas = vmap(area_polygon)(jnp.array(polygons))
    normal = vmap(normal_polygon)(jnp.array(polygons))

    print("Passed normals and areas tests\n")
    print("All good!")



from .artists import *  # noqa F403
from .plotters import *  # noqa F403
from .viewers import *  # noqa F403
from .notebooks import *  # noqa F403


__all__ = [name for name in dir() if not name.startswith('_')]



from jax_fdm.goals import VectorGoal

from jax_fdm.goals.node import NodeGoal


class NodePointGoal(VectorGoal, NodeGoal):
    """
    Make a node of a network to reach target xyz coordinates.
    """
    @staticmethod
    def prediction(eq_state, index):
        """
        The current xyz coordinates of the node in a network.
        """
        return eq_state.xyz[index, :]



import numpy as np

from jax import vmap

from jax_fdm import DTYPE_NP

from jax_fdm.goals import GoalState


# ==========================================================================
# Base goal
# ==========================================================================

class Goal:
    """
    The base goal.

    All goal subclasses must inherit from this class.
    """
    def __init__(self, key, target, weight):
        self._key = None
        self._weight = None
        self._target = None
        self._index = None

        self.key = key
        self.weight = weight
        self.target = target

        self.is_collectible = True

    @property
    def key(self):
        """
        The key of an element in a network.
        """
        return self._key

    @key.setter
    def key(self, key):
        if isinstance(key, int) or (isinstance(key, tuple) and len(key) == 2):
            key = key
        elif sum(isinstance(i, list) for i in key) > 0:
            key = key.pop()
        self._key = key

    @property
    def index(self):
        """
        The index of the goal key in the canonical ordering of a structure.
        """
        return self._index

    @index.setter
    def index(self, index):
        if isinstance(index, int):
            index = [index]
        self._index = np.array(index)

    @property
    def weight(self):
        """
        The importance of the goal.
        """
        return self._weight

    @weight.setter
    def weight(self, weight):
        self._weight = np.reshape(np.asarray(weight, dtype=DTYPE_NP), (-1, 1))

    @property
    def target(self):
        """
        The target to achieve.
        """
        raise NotImplementedError

    @staticmethod
    def goal(target, prediction):
        """
        The target to achieve.
        """
        return target

    def init(self, model, structure):
        """
        Initialize the goal with information from an equilibrium model.
        """
        self.index = self.index_from_model(model, structure)

    def __call__(self, eqstate):
        """
        Return the current goal state.
        """
        prediction = vmap(self.prediction, in_axes=(None, 0))(eqstate, self.index)
        goal = vmap(self.goal)(self.target, prediction)

        assert goal.shape == prediction.shape, f"Goal shape: {goal.shape} vs. Prediction shape: {prediction.shape}"

        return GoalState(goal=goal, prediction=prediction, weight=self.weight)


# ==========================================================================
# Base goal for a scalar quantity
# ==========================================================================

class ScalarGoal:
    """
    A goal that is expressed as a scalar quantity.
    """
    @property
    def target(self):
        """
        The target to achieve.
        """
        return self._target

    @target.setter
    def target(self, target):
        if isinstance(target, (int, float)):
            target = [target]
        self._target = np.reshape(np.array(target), (-1, 1))


# ==========================================================================
# Base goal for vector quantities
# ==========================================================================

class VectorGoal:
    """
    A goal that is expressed as a vector 3D quantity.
    """
    @property
    def target(self):
        """
        The target to achieve
        """
        return self._target

    @target.setter
    def target(self, target):
        self._target = np.reshape(np.asarray(target, dtype=DTYPE_NP), (-1, 3))



from jax_fdm.goals import Goal


class MeshGoal(Goal):
    """
    Base class for all goals that pertain to a subset of the nodes, faces and edges of a mesh.
    """
    def __init__(self, key=-1, target=0.0, weight=1.0):
        super().__init__(key=key, target=target, weight=weight)

    def index_from_model(self, model, structure):
        """
        The index of the goal key in a structure.
        """
        return -1



import numpy as np
import jax.numpy as jnp

from jax_fdm import DTYPE_JAX

from jax_fdm.parameters import split_parameters
from jax_fdm.parameters import combine_parameters

from jax_fdm.parameters import ParameterGroup

from jax_fdm.parameters import EdgeParameter
from jax_fdm.parameters import NodeSupportParameter
from jax_fdm.parameters import NodeLoadParameter

from jax_fdm.parameters import EdgeForceDensityParameter
from jax_fdm.parameters import NodeSupportXParameter
from jax_fdm.parameters import NodeSupportYParameter
from jax_fdm.parameters import NodeSupportZParameter
from jax_fdm.parameters import NodeLoadXParameter
from jax_fdm.parameters import NodeLoadYParameter
from jax_fdm.parameters import NodeLoadZParameter


class ParameterManager:
    """
    A parameter manager for optimization purposes.

    Parameters
    ----------
    model : `jax_fdm.equilibrium.EquilibriumModel`
        An equilibrium model.
    parameters : `List[jax_fdm.parameters.OptimizationParameter]`
        A list of optimization parameters.
    """
    parameter_types = [EdgeForceDensityParameter,
                       NodeSupportXParameter,
                       NodeSupportYParameter,
                       NodeSupportZParameter,
                       NodeLoadXParameter,
                       NodeLoadYParameter,
                       NodeLoadZParameter]

    def __init__(self, model, parameters, structure, network):
        """
        Initialize the manager.
        """
        self.model = model  # model.structure.network holds an FD network object.
        self.network = network
        self.structure = structure
        self.parameters = parameters

        self._indices_opt = None
        self._indices_frozen = None
        self._indices_optfrozen = None
        self._indices_groups = None

        self._indices_opt_sort = None
        self._indices_opt_unsort = None

        self._indices_fdm = None
        self._indices_fd = None
        self._indices_xyzfixed = None
        self._indices_loads = None

        self._parameters_value = None
        self._parameters_ordered = None
        self._parameters_model = None
        self._parameters_opt = None
        self._parameters_frozen = None

        self._bounds_low = None
        self._bounds_up = None

        self._startindex_fd = None
        self._startindex_xyzfixed = None
        self._startindex_loads = None

        self.init()

# ==========================================================================
# Warm start
# ==========================================================================

    def init(self):
        """
        Initialiaze the properties of this object so that every property becomes static after this call.

        TODO: This is fairly anti-pythonic. Please refactor me.
        """
        self.indices_fdm
        self.indices_optfrozen
        self.parameters
        self.parameters_model
        self.parameters_opt
        self.parameters_frozen
        self.parameters_ordered
        self.indices_groups
        self.indices_opt_unsort

# ==========================================================================
# Starting indices
# ==========================================================================

    @property
    def startindex_fd(self):
        """
        The starting index of the force density of the edges of a network.
        """
        if not self._startindex_fd:
            self._startindex_fd = 0
        return self._startindex_fd

    @property
    def startindex_xyzfixed(self):
        """
        The starting index of the xyz coordinates of the anchor nodes of a network.
        """
        if not self._startindex_xyzfixed:
            self._startindex_xyzfixed = self.structure.num_edges
        return self._startindex_xyzfixed

    @property
    def startindex_loads(self):
        """
        The starting index of the xyz coordinates of the loads at the nodes of a network.
        """
        if not self._startindex_loads:
            self._startindex_loads = self.structure.num_supports * 3 + self.startindex_xyzfixed
        return self._startindex_loads

# ==========================================================================
# Indices
# ==========================================================================

    @property
    def indices_fd(self):
        """
        The ordered indices of the force density of the edges of a network.
        """
        if self._indices_fd is None:
            start = self.startindex_fd
            stop = self.structure.num_edges
            self._indices_fd = np.array(range(start, start + stop))
        return self._indices_fd

    @property
    def indices_xyzfixed(self):
        """
        The ordered indices of the .
        """
        if self._indices_xyzfixed is None:
            start = self.startindex_xyzfixed
            stop = self.structure.num_supports * 3 + start
            self._indices_xyzfixed = np.array(range(start, stop))
        return self._indices_xyzfixed

    @property
    def indices_loads(self):
        """
        The ordered indices of the xyz coordinates of the anchor nodes of a network.
        """
        if self._indices_loads is None:
            start = self.startindex_loads
            stop = self.structure.num_nodes * 3 + start
            self._indices_loads = np.array(range(start, stop))
        return self._indices_loads

    @property
    def indices_groups(self):
        """
        A list with indices distributions optimization parameters to parameter groups.
        """
        if self._indices_groups is None:
            indices = []
            for idx, parameter in enumerate(self.parameters_ordered):
                if isinstance(parameter, ParameterGroup):
                    for j in range(len(parameter.key)):
                        indices.append(idx)
                else:
                    indices.append(idx)

            self._indices_groups = np.array(indices, dtype=np.int64)
            # self._indices_groups = tuple(indices)

        return self._indices_groups

    @property
    def indices_opt(self):
        """
        The type-ordered indices of the optimization parameters.
        """
        if self._indices_opt is None:
            indices = []
            pshift = 0
            # iterate over ordered parameter classes typed
            for ptype in self.parameter_types:
                # get the indices of the all the parameters of this parameter class
                itype = self._indices_type(ptype)
                # shift class indices by class shift
                itype_shifted = self._indices_shift(itype, pshift)
                # store shifted class indices in indices list
                indices.extend(itype_shifted)
                # update class shift
                pshift += self._shift_type(ptype)

            self._indices_opt = indices
            self._indices_opt = np.array(indices, dtype=np.int64)

        return self._indices_opt

    @property
    def indices_opt_sort(self):
        """
        The indices that sort the index-based ordering of the type-sorted optimization parameters.
        """
        if self._indices_opt_sort is None:
            indices = np.argsort(self.indices_opt)
            self._indices_opt_sort = indices
        return self._indices_opt_sort

    @property
    def indices_opt_unsort(self):
        """
        The indices that unsort the index-based ordering of the type-sorted optimization parameters.
        """
        if self._indices_opt_unsort is None:
            indices = np.argsort(self.indices_opt_sort)
            self._indices_opt_unsort = indices
        return self._indices_opt_unsort

    @property
    def indices_optfrozen(self):
        if self._indices_optfrozen is None:
            _, indices = split_parameters(self.parameters_model, func=self.mask_optimizable)
            self._indices_optfrozen = indices
        return self._indices_optfrozen

    @property
    def indices_fdm(self):
        if self._indices_fdm is None:
            self._indices_fdm = [self.startindex_xyzfixed, self.startindex_loads]
        return self._indices_fdm

# ==========================================================================
# Helpers
# ==========================================================================

    def _indices_type(self, cls):
        """
        Compute the parameter index if a parameter is an instance of a given type.
        """
        indices = []
        for parameter in self.parameters:
            if isinstance(parameter, cls):
                if isinstance(parameter, ParameterGroup):
                    indices.extend(parameter.index(self.model, self.structure))
                else:
                    indices.append(parameter.index(self.model, self.structure))

        # return np.array(indices, dtype=np.int64)
        return indices

    def _indices_shift(self, indices, shift):
        """
        Shift a collection of parameter indices by a given integer.
        """
        return [idx + shift for idx in indices]

    def _shift_type(self, ptype):
        """
        The number of indices to shift of a collection of parameters of a given type.
        """
        if issubclass(ptype, EdgeParameter):
            shift = self.structure.num_edges
        elif issubclass(ptype, NodeSupportParameter):
            shift = self.structure.num_supports
        elif issubclass(ptype, NodeLoadParameter):
            shift = self.structure.num_nodes

        return shift

# ==========================================================================
# Bounds
# ==========================================================================

    @property
    def bounds_low(self):
        """
        Return an array with the lower bound of the optimization parameters.
        """
        if self._bounds_low is None:
            bounds = []
            for parameter in self.parameters_ordered:
                bounds.append(parameter.bound_low)

            self._bounds_low = np.array(bounds)

        return self._bounds_low

    @property
    def bounds_up(self):
        """
        Return an array with the upper bound of the optimization parameters.
        """
        if self._bounds_up is None:
            bounds = []
            for parameter in self.parameters_ordered:
                bounds.append(parameter.bound_up)

            self._bounds_up = np.array(bounds)

        return self._bounds_up

    @property
    def parameters_value(self):
        """
        Return an array with the intial value of the optimization parameters.
        """
        if self._parameters_value is None:
            values = []
            for parameter in self.parameters_ordered:
                values.append(parameter.value(self.model, self.network))
            self._parameters_value = jnp.array(values, dtype=DTYPE_JAX)

        return self._parameters_value

# ==========================================================================
# Parameters
# ==========================================================================

    @property
    def parameters_ordered(self):
        """
        The optimization parameter objects, sorted by type.
        """
        if self._parameters_ordered is None:

            parameters = []
            for ptype in self.parameter_types:

                for parameter in self.parameters:
                    if isinstance(parameter, ptype):
                        parameters.append(parameter)

            assert len(parameters) > 0, "No valid optimization parameters defined!"

            self._parameters_ordered = parameters

        return self._parameters_ordered

    @property
    def parameters_model(self):
        """
        The model parameters as a single array.
        """
        if self._parameters_model is None:
            q, xyz_fixed, loads = [jnp.asarray(p, dtype=DTYPE_JAX) for p in self.network.parameters()]
            self._parameters_model = jnp.concatenate((q,
                                                      jnp.ravel(xyz_fixed, order="F"),
                                                      jnp.ravel(loads, order="F")))
        return self._parameters_model

    @property
    def parameters_opt(self):
        """
        The optimizable model parameters.
        """
        if self._parameters_opt is None:
            parameters, _ = split_parameters(self.parameters_model, func=self.mask_optimizable)
            opt, _ = parameters
            self._parameters_opt = opt
        return self._parameters_opt

    @property
    def parameters_frozen(self):
        """
        The non-optimizable model parameters.
        """
        if self._parameters_frozen is None:
            parameters, _ = split_parameters(self.parameters_model, func=self.mask_optimizable)
            _, frozen = parameters
            self._parameters_frozen = frozen
        return self._parameters_frozen

    def parameters_fdm(self, params_opt):
        """
        Convert optimization parameters into fdm parameters.
        """
        # unsort params opt
        params_opt = params_opt[self.indices_groups]
        params_opt = params_opt[self.indices_opt_sort]
        params = combine_parameters((params_opt, self.parameters_frozen),
                                    adef=self.indices_optfrozen)
        q, xyz_fixed, loads = jnp.split(params, self.indices_fdm)

        return q, jnp.reshape(xyz_fixed, (-1, 3), order="F"), jnp.reshape(loads, (-1, 3), order="F")

# ==========================================================================
# Masks
# ==========================================================================

    def mask_optimizable(self, array):
        """
        Returns two boolean masks.

        One with ones denoting the optimizable parameters.
        The second one denoting the frozen parameters.
        """
        mask = np.zeros_like(array, dtype=np.int64)
        mask[self.indices_opt] = 1

        return mask, np.logical_not(mask)

    def mask_fdm(self, array):
        """
        """
        for indices in (self.indices_fd,
                        self.indices_xyzfixed,
                        self.indices_loads):
            mask = np.zeros_like(array, dtype=np.int64)
            mask[indices] = 1
            yield mask



import jax
import jax.numpy as jnp

from typing import Union
from typing import NamedTuple

from jax_fdm.datastructures import FDNetwork
from jax_fdm.datastructures import FDMesh


# ==========================================================================
# Equilibrium state
# ==========================================================================

class EquilibriumState(NamedTuple):
    xyz: jax.Array
    residuals: jax.Array
    lengths: jax.Array
    forces: jax.Array
    loads: jax.Array
    vectors: jax.Array


# ==========================================================================
# Load state
# ==========================================================================

class LoadState(NamedTuple):
    nodes: Union[jax.Array, float]
    edges: Union[jax.Array, float]
    faces: Union[jax.Array, float]

    @classmethod
    def from_datastructure(cls, datastructure, dtype=None):
        """
        Create a load state from a datastructure.
        """
        loads_edges = jnp.asarray(datastructure.edges_loads(), dtype)
        if jnp.allclose(loads_edges, 0.0):
            loads_edges = 0.0

        if isinstance(datastructure, FDNetwork):
            loads_nodes = jnp.asarray(datastructure.nodes_loads())
            loads_faces = 0.0

        elif isinstance(datastructure, FDMesh):
            loads_nodes = jnp.asarray(datastructure.vertices_loads(), dtype)

            loads_faces = jnp.asarray(datastructure.faces_loads(), dtype)
            if jnp.allclose(loads_faces, 0.0):
                loads_faces = 0.0

        return cls(nodes=loads_nodes,
                   edges=loads_edges,
                   faces=loads_faces)


# ==========================================================================
# Parameter state
# ==========================================================================

class EquilibriumParametersState(NamedTuple):
    q: jax.Array
    xyz_fixed: jax.Array
    loads: LoadState

    @classmethod
    def from_datastructure(cls, datastructure, dtype=None):
        """
        Create a parameter state from a datastructure.
        """
        q = datastructure.edges_forcedensities()

        if isinstance(datastructure, FDNetwork):
            xyz_fixed = datastructure.nodes_fixedcoordinates()

        elif isinstance(datastructure, FDMesh):
            xyz_fixed = datastructure.vertices_fixedcoordinates()

        return cls(q=jnp.asarray(q, dtype),
                   xyz_fixed=jnp.asarray(xyz_fixed, dtype),
                   loads=LoadState.from_datastructure(datastructure))



from .states import *  # noqa F403
from .structures import *  # noqa F403
from .loads import *  # noqa F403
from .sparse import *  # noqa F403
from .solvers import *  # noqa F403
from .models import *  # noqa F403
from .fdm import *  # noqa F403


__all__ = [name for name in dir() if not name.startswith('_')]



from compas.artists import Artist

from jax_fdm.datastructures import FDNetwork

from jax_fdm.visualization.plotters import FDNetworkPlotterArtist
from jax_fdm.visualization.plotters import FDVector
from jax_fdm.visualization.plotters import FDVectorPlotterArtist


__all__ = ["register_artists"]


def register_artists():
    """
    Register objects to the artist factory.
    """
    Artist.register(FDVector, FDVectorPlotterArtist, context="Plotter")
    Artist.register(FDNetwork, FDNetworkPlotterArtist, context="Plotter")



from math import fabs

from collections.abc import Iterable

from compas.artists import NetworkArtist

from compas.colors import Color
from compas.colors import ColorMap

from compas.utilities import remap_values


__all__ = ["FDNetworkArtist"]


class FDNetworkArtist(NetworkArtist):
    """
    The base artist to display a force density network across different contexts.
    """
    default_edgecolor = Color.teal()
    default_nodecolor = Color.grey().lightened(factor=100)
    default_nodesupportcolor = Color.from_rgb255(0, 150, 10)

    default_loadcolor = Color.from_rgb255(0, 150, 10)
    default_reactioncolor = Color.pink()

    default_tensioncolor = Color.from_rgb255(227, 6, 75)
    default_compressioncolor = Color.from_rgb255(12, 119, 184)

    default_fdcolormap = ColorMap.from_mpl("viridis")
    default_forcecolormap = ColorMap.from_three_colors(Color.from_rgb255(12, 119, 184),
                                                       Color.grey().lightened(50),
                                                       Color.from_rgb255(227, 6, 75))

    default_nodesize = 0.1
    default_edgewidth = (0.01, 0.1)
    default_loadscale = 1.0
    default_reactionscale = 1.0
    default_loadtol = 1e-3
    default_reactiontol = 1e-3

    def __init__(self,
                 network,
                 nodes=None,
                 edges=None,
                 nodecolor=None,
                 edgecolor=None,
                 nodesize=None,
                 edgewidth=None,
                 loadcolor=None,
                 loadscale=None,
                 loadtol=None,
                 reactioncolor=None,
                 reactionscale=None,
                 reactiontol=None,
                 show_nodes=False,
                 show_edges=True,
                 show_loads=True,
                 show_reactions=True,
                 show_supports=True,
                 *args,
                 **kwargs):
        super().__init__(network=network,
                         nodes=nodes,
                         edges=edges,
                         nodecolor=nodecolor,
                         *args,
                         **kwargs)

        self._default_loadcolor = None
        self._default_reactioncolor = None
        self._default_nodesupportcolor = None
        self._default_tensioncolor = None
        self._default_compressioncolor = None

        self._node_size = None

        self.node_size = nodesize
        self.edge_width = edgewidth

        self.edge_color = edgecolor

        self.load_color = loadcolor or self.default_loadcolor

        if edgecolor == "force":
            self.default_reactioncolor = Color.from_rgb255(0, 150, 10)
        self.reaction_color = reactioncolor or self.default_reactioncolor

        self.load_scale = loadscale or self.default_loadscale
        self.load_tol = loadtol or self.default_loadtol

        self.reaction_scale = reactionscale or self.default_reactionscale
        self.reaction_tol = reactiontol or self.default_reactiontol

        self.show_nodes = show_nodes
        self.show_edges = show_edges
        self.show_loads = show_loads
        self.show_reactions = show_reactions
        self.show_supports = show_supports

        self.collection_edges = None
        self.collection_nodes = None
        self.collection_loads = None
        self.collection_reactions = None

        self._init_edgecolor = edgecolor
        self._init_edgewidth = edgewidth

    # ==========================================================================
    # Draw
    # ==========================================================================

    def draw(self):
        """
        Draw everything.
        """
        if self.show_edges:
            self.collection_edges = self.draw_edges()
        if self.show_nodes:
            self.collection_nodes = self.draw_nodes()
        if self.show_loads:
            self.collection_loads = self.draw_loads()
        if self.show_reactions:
            self.collection_reactions = self.draw_reactions()

    # ==========================================================================
    # Draw collections
    # ==========================================================================

    def draw_nodes(self):
        """
        Draw the nodes of the network.
        """
        nodes = {}
        for node in self.nodes:
            size = self.node_size[node]
            color = self.node_color[node]
            nodes[node] = self.draw_node(node, size, color)

        return nodes

    def draw_edges(self):
        """
        Draw the edges of the network.
        """
        edges = {}

        for edge in self.edges:
            width = self.edge_width[edge]
            color = self.edge_color[edge]
            edges[edge] = self.draw_edge(edge, width, color)

        return edges

    def draw_loads(self):
        """
        Draw the loads at the nodes of the network.
        """
        loads = {}

        for node in self.nodes:
            load = self.draw_load(node, self.load_scale, self.load_color)
            if load:
                loads[node] = load

        return loads

    def draw_reactions(self):
        """
        Draw the reactions at the nodes of the network.
        """
        reactions = {}

        for node in self.nodes:
            reaction = self.draw_reaction(node, self.reaction_scale, self.reaction_color)
            if reaction:
                reactions[node] = reaction

        return reactions

    # ==========================================================================
    # Draw elements
    # ==========================================================================

    def draw_node(self, node, size, color):
        """
        Draw a node.
        """
        raise NotImplementedError

    def draw_edge(self, edge, width, color):
        """
        Draw an edge.
        """
        raise NotImplementedError

    def draw_load(self, node, scale, color):
        """
        Draw a load.
        """
        raise NotImplementedError

    def draw_reaction(self, node, scale, color):
        """
        Draw a load.
        """
        raise NotImplementedError

    def clear_nodes(self):
        """
        Clear the nodes.
        """
        pass

    def clear_edges(self):
        """
        Clear the edges.
        """
        pass

    # ==========================================================================
    # Update elements
    # ==========================================================================

    def update(self, eqstate):
        """
        Update the attributes of the network based on an equilibrium state.
        """
        raise NotImplementedError

    # ==========================================================================
    # Properties
    # ==========================================================================

    @property
    def edge_color(self):
        """
        The edge colors.
        """
        if not self._edge_color:
            self._edge_color = {edge: self.default_edgecolor for edge in self.edges}
        return self._edge_color

    @edge_color.setter
    def edge_color(self, color):
        if isinstance(color, dict):
            self._edge_color = color

        elif isinstance(color, Color):
            self._edge_color = {edge: color for edge in self.edges}

        elif isinstance(color, str):
            network = self.network

            if color == "fd":
                cmap = self.default_fdcolormap
                values = [fabs(network.edge_forcedensity(edge)) for edge in self.edges]
                try:
                    ratios = remap_values(values)
                except ZeroDivisionError:
                    ratios = [0.0] * len(self.edges)

                self._edge_color = {edge: cmap(ratio) for edge, ratio in zip(self.edges, ratios)}

            elif color == "force":
                cmap = self.default_forcecolormap

                edge_color = {}
                for edge in self.edges:
                    force = network.edge_force(edge)
                    _color = self.default_tensioncolor
                    if force <= 0.0:
                        _color = self.default_compressioncolor
                    edge_color[edge] = _color

                self._edge_color = edge_color

    @property
    def node_color(self):
        """
        The node colors.
        """
        if self._node_color:
            return self._node_color

        node_color = {}

        for node in self.nodes:
            color = self.default_nodecolor
            if self.network.node_attribute(node, "is_support"):
                color = self.default_nodesupportcolor
            node_color[node] = color

        self._node_color = node_color

        return self._node_color

    @node_color.setter
    def node_color(self, color):
        if isinstance(color, dict):
            self._node_color = color
        elif isinstance(color, Color):
            self._node_color = {node: color for node in self.nodes}

    @property
    def edge_width(self):
        """
        The width of the edges.
        """
        if not self._edge_width:
            self.edge_width = self.default_edgewidth
        return self._edge_width

    @edge_width.setter
    def edge_width(self, width):
        if isinstance(width, dict):
            self._edge_width = width

        elif isinstance(width, (int, float)):
            self._edge_width = {edge: width for edge in self.edges}

        elif isinstance(width, Iterable) and len(width) == 2:
            width_min, width_max = width
            forces = [fabs(self.network.edge_force(edge)) for edge in self.edges]

            if min(forces) == max(forces):
                widths = [width_max] * len(self.edges)
            else:
                try:
                    widths = remap_values(forces, width_min, width_max)
                except ZeroDivisionError:
                    widths = [self.default_edgewidth][0] * len(self.edges)

            self._edge_width = {edge: width for edge, width in zip(self.edges, widths)}



import jax.numpy as jnp

from jax import vmap

from jax_fdm.goals import ScalarGoal

from jax_fdm.goals.network import NetworkGoal


class NetworkSmoothGoal(ScalarGoal, NetworkGoal):
    """
    Smudge a network based on the fairness of its nodes.

    Notes
    -----
    Fairness is an energy that measures the smoothness of the
    position of a node w.r.t. to that of its neighbors.

    The fairness energy is computed as the squared length of the
    vector between every node position and their neighbors' centroid.
    """
    def __init__(self):
        super().__init__()
        self.adjacency = None
        self.indices_free = None

    def prediction(self, eq_state, index):
        """
        The current fairness value of the node.
        """
        xyz = eq_state.xyz[self.indices_free, :]
        adjacency = self.adjacency[self.indices_free, :]
        fairness_fn = vmap(node_nbrs_fairness_ngon, in_axes=(None, 0, 0))

        fairness_nodes = fairness_fn(eq_state.xyz, xyz, adjacency)
        fairness = jnp.sum(fairness_nodes)

        return jnp.atleast_1d(fairness)

    def init(self, model, structure):
        """
        Initialize the constraint with information from an equilibrium model.
        """
        super().init(model, structure)
        self.indices_free = structure.indices_free
        self.adjacency = structure.adjacency


def node_nbrs_fairness_ngon(xyz_all, xyz_node, adjacency_node):
    """
    Compute the fairness of an n-gon node neighborhood.
    """
    xyz_nbrs = adjacency_node @ xyz_all / jnp.sum(adjacency_node, axis=-1)

    fvector = xyz_node - xyz_nbrs
    assert fvector.shape == xyz_node.shape

    return jnp.sum(jnp.square(fvector))



"""
A force density mesh.
"""
import numpy as np

from compas.datastructures import Mesh

from jax_fdm.datastructures import FDDatastructure

from jax_fdm.geometry import polygon_lcs


class FDMesh(Mesh, FDDatastructure):
    """
    A force density mesh.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.update_default_edge_attributes({"q": 0.0,
                                             "length": 0.0,
                                             "force": 0.0,
                                             "px": 0.0,
                                             "py": 0.0,
                                             "pz": 0.0})

        self.update_default_vertex_attributes({"x": 0.0,
                                               "y": 0.0,
                                               "z": 0.0,
                                               "px": 0.0,
                                               "py": 0.0,
                                               "pz": 0.0,
                                               "rx": 0.0,
                                               "ry": 0.0,
                                               "rz": 0.0,
                                               "is_support": False})

        self.update_default_face_attributes({"px": 0.0,
                                             "py": 0.0,
                                             "pz": 0.0})

    # ----------------------------------------------------------------------
    # Node
    # ----------------------------------------------------------------------

    def is_vertex_support(self, key):
        """
        Test if the vertex is a support.
        """
        return self.vertex_attribute(key, name="is_support")

    def number_of_supports(self):
        """
        The number of supported vertices.
        """
        return len(list(self.vertices_supports()))

    def vertex_support(self, key):
        """
        Sets a vertex to a fixed support.
        """
        return self.vertex_attribute(key, name="is_support", value=True)

    def vertex_load(self, key, load=None):
        """
        Gets or sets a load to a vertex.
        """
        return self.vertex_attributes(key, names=("px", "py", "pz"), values=load)

    def vertex_residual(self, key):
        """
        Gets the residual force of a mesh vertex.
        """
        return self.vertex_attributes(key, names=("rx", "ry", "rz"))

    def vertex_reaction(self, key):
        """
        Gets the reaction force of a mesh vertex.
        """
        return self.vertex_residual(key)

    def vertices_coordinates(self, keys=None, axes="xyz"):
        """
        Gets or sets the x, y, z coordinates of a list of vertices.
        """
        keys = keys or self.vertices()
        return [self.vertex_coordinates(node, axes) for node in keys]

    def vertices_fixedcoordinates(self, keys=None, axes="xyz"):
        """
        Gets the x, y, z coordinates of the supports of the network.
        """
        if keys:
            keys = {key for key in keys if self.is_vertex_support(key)}
        else:
            keys = self.vertices_fixed()

        return [self.vertex_coordinates(node, axes) for node in keys]

    def vertices_supports(self, keys=None):
        """
        Gets or sets the vertex keys where a support has been assigned.
        """
        if keys is None:
            return self.vertices_where({"is_support": True})

        return self.vertices_attribute(name="is_support", value=True, keys=keys)

    def vertices_fixed(self, keys=None):
        """
        Gets or sets the vertex keys where a support has been assigned.
        """
        return self.vertices_supports(keys)

    def vertices_free(self):
        """
        The keys of the vertices where there is no support assigned.
        """
        return self.vertices_where({"is_support": False})

    def vertices_loads(self, load=None, keys=None):
        """
        Gets or sets a load to the vertices of the mesh.
        """
        return self.vertices_attributes(names=("px", "py", "pz"), values=load, keys=keys)

    def vertices_residual(self, keys=None):
        """
        Gets the residual forces at the vertices of the mesh.
        """
        return self.vertices_attributes(names=("rx", "ry", "rz"), keys=keys)

    def vertices_reactions(self, keys=None):
        """
        Gets the reaction forces at the vertices of the mesh.
        """
        keys = keys or self.vertices_fixed()
        return self.vertices_residual(keys)

    # ----------------------------------------------------------------------
    # Edges
    # ----------------------------------------------------------------------

    def is_edge_supported(self, key):
        """
        Test if any of edge vertices is a support.
        """
        return any(self.is_vertex_support(vertex) for vertex in key)

    def is_edge_fully_supported(self, key):
        """
        Test if all the edge vertices are a support.
        """
        return all(self.is_vertex_support(vertex) for vertex in key)

    # ----------------------------------------------------------------------
    # Faces
    # ----------------------------------------------------------------------

    def face_lcs(self, key):
        """
        Calculate the local coordinate system (LCS) of this face.
        """
        return polygon_lcs(np.asarray(self.face_coordinates(key))).tolist()

    def face_load(self, key, load=None):
        """
        Gets or sets a load on a face.
        """
        return self.face_attributes(key=key, names=("px", "py", "pz"), values=load)

    def is_face_supported(self, key):
        """
        Test if any of the face vertices is a support.
        """
        return any(self.is_vertex_support(vertex) for vertex in self.face_vertices(key))

    def is_face_fully_supported(self, key):
        """
        Test if all the face vertices are a support.
        """
        return all(self.is_vertex_support(vertex) for vertex in self.face_vertices(key))

    def faces_loads(self, load=None, keys=None):
        """
        Gets or sets a load on the faces.
        """
        return self.faces_attributes(names=("px", "py", "pz"), values=load, keys=keys)

    # ----------------------------------------------------------------------
    # Datastructure properties
    # ----------------------------------------------------------------------

    def parameters(self):
        """
        Return the design parameters of the network.
        """
        q = self.edges_forcedensities()
        xyz_fixed = self.vertices_fixedcoordinates()
        loads = self.vertices_loads()

        return q, xyz_fixed, loads

    # ----------------------------------------------------------------------
    # Maps
    # ----------------------------------------------------------------------

    def index_uv(self):
        """
        Returns a dictionary that maps edges in a list to the corresponding vertex key pairs.
        """
        return dict(enumerate(self.edges()))

    def uv_index(self):
        """
        Returns a dictionary that maps edge keys (i.e. pairs of vertex keys)
        to the corresponding edge index in a list or array of edges.
        """
        return {(u, v): index for index, (u, v) in enumerate(self.edges())}



from jax_fdm.goals import ScalarGoal

from jax_fdm.goals.node import NodeGoal


class NodeXCoordinateGoal(ScalarGoal, NodeGoal):
    """
    Make a node of a network to reach a target X coordinate.
    """
    @staticmethod
    def prediction(eq_state, index):
        """
        The current X coordinate of the node in a network.
        """
        return eq_state.xyz[index, :1]


class NodeYCoordinateGoal(ScalarGoal, NodeGoal):
    """
    Make a node of a network to reach a target Y coordinate.
    """
    @staticmethod
    def prediction(eq_state, index):
        """
        The current Y coordinate of the node in a network.
        """
        return eq_state.xyz[index, 1:2]


class NodeZCoordinateGoal(ScalarGoal, NodeGoal):
    """
    Make a node of a network to reach a target Z coordinate.
    """
    @staticmethod
    def prediction(eq_state, index):
        """
        The current Z coordinate of the node in a network.
        """
        return eq_state.xyz[index, 2:]



from jax_fdm.constraints import Constraint


class NetworkConstraint(Constraint):
    """
    Base class for all constraints that pertain to all the edges or all the nodes of a network.
    """
    def __init__(self, bound_low, bound_up):
        super().__init__(key=-1, bound_low=bound_low, bound_up=bound_up)

    def index_from_model(self, model, structure):
        """
        Method only for API compatibility.
        """
        return -1



from jax_fdm.constraints.edge import EdgeConstraint


class EdgeLengthConstraint(EdgeConstraint):
    """
    Constraints the length of an edge between a lower and an upper bound.
    """
    @staticmethod
    def constraint(eqstate, index):
        """
        Returns the length of an edge from an equilibrium state.
        """
        return eqstate.lengths[index, :]



import numpy as np

import jax
import jax.numpy as jnp

from jax.experimental.sparse import BCOO
from jax.experimental.sparse import CSC

from jax_fdm import DTYPE_INT_JAX
from jax_fdm import DTYPE_INT_NP

from jax_fdm.equilibrium.structures.topology import Graph
from jax_fdm.equilibrium.structures.topology import GraphSparse
from jax_fdm.equilibrium.structures.topology import Mesh
from jax_fdm.equilibrium.structures.topology import MeshSparse


# ==========================================================================
# Structure
# ==========================================================================

class EquilibriumStructure(Graph):
    """
    A structure.
    """
    supports: np.ndarray

    connectivity_free: jax.Array
    connectivity_fixed: jax.Array

    indices_free: jax.Array
    indices_fixed: jax.Array
    indices_freefixed: jax.Array

    def __init__(self, nodes, edges, supports, **kwargs):
        super().__init__(nodes=nodes, edges=edges, **kwargs)

        self.supports = supports

        self.indices_free = self._indices_free()
        self.indices_fixed = self._indices_fixed()
        self.indices_freefixed = self._indices_freefixed()

        self.connectivity_free = self._connectivity_free()
        self.connectivity_fixed = self._connectivity_fixed()

    @classmethod
    def from_network(cls, network):
        """
        Create a structure from a force density network.
        """
        nodes = list(network.nodes())
        edges = list(network.edges())

        supports = []
        for node in nodes:
            flag = 0.0
            if network.is_node_support(node):
                flag = 1.0
            supports.append(flag)

        nodes = np.asarray(nodes, dtype=DTYPE_INT_NP)
        edges = np.asarray(edges, dtype=DTYPE_INT_NP)
        supports = np.asarray(supports, dtype=DTYPE_INT_NP)

        return cls(nodes, edges, supports)

    @property
    def num_supports(self):
        """
        The number of supports.
        """
        return jnp.count_nonzero(self.supports)

    @property
    def num_free(self):
        """
        The number of supports.
        """
        return self.num_nodes - self.num_supports

    @property
    def support_index(self):
        """
        A mapping from support keys to indices.
        """
        return {int(key): index for index, key in enumerate(self.nodes_fixed)}

    @property
    def nodes_free(self):
        """
        The free nodes.
        """
        return self.nodes[self.indices_free]

    @property
    def nodes_fixed(self):
        """
        The fixed nodes.
        """
        return self.nodes[self.indices_fixed]

    def _connectivity_free(self):
        """
        The connectivity matrix between edges and nodes.
        """
        return self.connectivity[:, self.indices_free]

    def _connectivity_fixed(self):
        """
        The connectivity matrix between edges and nodes.
        """
        return self.connectivity[:, self.indices_fixed]

    def _indices_free(self):
        """
        The indices of the unsupported nodes in the structure.
        """
        indices = jnp.flatnonzero(self.supports == 0, size=self.num_free)

        return indices

    def _indices_fixed(self):
        """
        The indices of the unsupported nodes in the structure.
        """
        indices = jnp.flatnonzero(self.supports, size=self.num_supports)

        return indices

    def _indices_freefixed(self):
        """
        A list with the node keys of all the nodes sorted by their node index.
        """
        # TODO: this method must be refactored to be more transparent.
        freefixed_indices = jnp.concatenate([self.indices_free,
                                             self.indices_fixed])

        indices = {node.item(): index for index, node in enumerate(freefixed_indices)}
        sorted_indices = []
        for _, index in sorted(indices.items(), key=lambda item: item[0]):
            sorted_indices.append(index)

        return jnp.asarray(sorted_indices, dtype=DTYPE_INT_JAX)


# ==========================================================================
# Sparse structure
# ==========================================================================

class EquilibriumStructureSparse(EquilibriumStructure, GraphSparse):
    """
    A sparse structure.
    """
    diag_indices: jax.Array
    index_array: jax.Array
    diags: jax.Array

    def __init__(self, nodes, edges, supports, **kwargs):
        super().__init__(nodes=nodes,
                         edges=edges,
                         supports=supports,
                         **kwargs)

        # Do some precomputation to be able to construct
        # the lhs matrix through indexing
        c_free_csc = self.connectivity_scipy[:, self.indices_free]
        index_array = self._get_sparse_index_array(c_free_csc)
        self.index_array = index_array

        # Indices of data corresponding to diagonal.
        # With this array we can just index directly into the
        # CSC.data array to refer to the diagonal entries.
        self.diag_indices = self._get_sparse_diag_indices(index_array)

        # Prepare the array D st when D.T @ q we get the diagonal elements of matrix
        self.diags = self._get_sparse_diag_data(c_free_csc)

    def _connectivity_free(self):
        """
        The connectivity matrix between edges and nodes.
        """
        return BCOO.from_scipy_sparse(self.connectivity_scipy[:, self.indices_free])

    def _connectivity_fixed(self):
        """
        The connectivity matrix between edges and nodes.
        """
        return BCOO.from_scipy_sparse(self.connectivity_scipy[:, self.indices_fixed])

    @staticmethod
    def _get_sparse_index_array(c_free_csc):
        """
        Create an index array such that the off-diagonals can index into the
        force density vector.

        This array is used to create the off-diagonal entries of the lhs matrix.

        # NOTE: The input matrix must be a scipy sparse array!
        """
        fd_mod_c_free_csc = c_free_csc.copy()
        fd_mod_c_free_csc.data *= np.take(np.arange(c_free_csc.shape[0]) + 1,
                                          c_free_csc.indices)
        index_array = -(c_free_csc.T @ fd_mod_c_free_csc)

        # The diagonal entries should be set to 0 so that it indexes
        # into a valid entry, but will later be overwritten.
        index_array.setdiag(0)
        index_array = index_array.astype(int)

        return CSC((index_array.data, index_array.indices, index_array.indptr),
                   shape=index_array.shape)

    @staticmethod
    def _get_sparse_diag_indices(csc):
        """
        Given a CSC matrix, get indices into `data` that access diagonal elements in order.
        """
        all_indices = []
        for i in range(csc.shape[0]):
            index_range = csc.indices[csc.indptr[i]:csc.indptr[i + 1]]
            ind_loc = jnp.where(index_range == i)[0]
            all_indices.append(ind_loc + csc.indptr[i])

        return jnp.concatenate(all_indices)

    @staticmethod
    def _get_sparse_diag_data(c_free_csc):
        """
        The diagonal of the lhs matrix is the sum of force densities for
        each outgoing/incoming edge on the node.

        We create the `diags` matrix such that when we multiply it with the
        force density vector we get the diagonal.
        """
        diags_data = jnp.ones_like(c_free_csc.data)

        return CSC((diags_data, c_free_csc.indices, c_free_csc.indptr),
                   shape=c_free_csc.shape)


# ==========================================================================
# Mesh structures
# ==========================================================================

class EquilibriumMeshStructure(EquilibriumStructure, Mesh):
    """
    An equilibrium mesh structure.
    """
    def __init__(self, vertices, faces, edges, supports, **kwargs):
        super().__init__(nodes=vertices,
                         edges=edges,
                         supports=supports,
                         vertices=vertices,
                         faces=faces,
                         **kwargs)

    @property
    def num_free(self):
        """
        The number of supports.
        """
        return self.num_vertices - self.num_supports

    @classmethod
    def from_mesh(cls, mesh):
        """
        Create a structure from a force density mesh.
        """
        vertices = list(mesh.vertices())
        edges = list(mesh.edges())

        supports = []
        for vertex in vertices:
            flag = 0.0
            if mesh.is_vertex_support(vertex):
                flag = 1.0
            supports.append(flag)

        faces = [mesh.face_vertices(fkey) for fkey in mesh.faces()]
        max_length_face = max(len(face) for face in faces)

        pad_value = -1
        padded_faces = []
        for face in faces:
            len_face = len(face)
            if len_face < max_length_face:
                face = face + [pad_value] * (max_length_face - len_face)
            padded_faces.append(face)

        faces = np.asarray(padded_faces, dtype=DTYPE_INT_NP)
        vertices = np.asarray(vertices, dtype=DTYPE_INT_NP)
        edges = np.asarray(edges, dtype=DTYPE_INT_NP)
        supports = np.asarray(supports, dtype=DTYPE_INT_NP)

        return cls(vertices, faces, edges, supports)

    @property
    def support_index(self):
        """
        A mapping from support vertices keys to indices.
        """
        return {int(key): index for index, key in enumerate(self.vertices_fixed)}

    @property
    def vertices_free(self):
        """
        The free vertices.
        """
        return self.vertices[self.indices_free]

    @property
    def vertices_fixed(self):
        """
        The fixed vertices.
        """
        return self.vertices[self.indices_fixed]


class EquilibriumMeshStructureSparse(EquilibriumMeshStructure, EquilibriumStructureSparse, MeshSparse):
    """
    An equilibrium mesh structure.
    """
    def __init__(self, vertices, faces, edges, supports, **kwargs):
        super().__init__(vertices=vertices,
                         faces=faces,
                         edges=edges,
                         supports=supports,
                         **kwargs)



"""
A bunch of goals to strive for.
"""
from jax_fdm.geometry import length_vector
from jax_fdm.geometry import normalize_vector

from jax_fdm.goals import ScalarGoal
from jax_fdm.goals import VectorGoal

from jax_fdm.goals.node import NodeGoal


class NodeResidualForceGoal(ScalarGoal, NodeGoal):
    """
    Make the residual force in a network to match a non-negative magnitude.
    """
    @staticmethod
    def prediction(eq_state, index):
        """
        The residual at the the predicted node of the network.
        """
        residual = eq_state.residuals[index, :]

        return length_vector(residual)


class NodeResidualVectorGoal(VectorGoal, NodeGoal):
    """
    Make the residual force in a network to match the magnitude and direction of a vector.
    """
    @staticmethod
    def prediction(eq_state, index):
        """
        The residual at the the predicted node of the network.
        """
        return eq_state.residuals[index, :]


class NodeResidualDirectionGoal(VectorGoal, NodeGoal):
    """
    Make the residual force in a network to match the direction of a vector.

    Another effective proxy for cosine distance can be obtained by
    L2 normalisation of the vectors, followed by the application of normal
    Euclidean distance. Using this technique each term in each vector is
    first divided by the magnitude of the vector, yielding a vector of unit
    length. Then, it is clear, the Euclidean distance over the end-points
    of any two vectors is a proper metric which gives the same ordering as
    the cosine distance (a monotonic transformation of Euclidean distance;
    see below) for any comparison of vectors, and furthermore avoids the
    potentially expensive trigonometric operations required to yield
    a proper metric.
    """
    @staticmethod
    def prediction(eq_state, index):
        """
        The residual at the the predicted node of the network.
        """
        residual = eq_state.residuals[index, :]

        return normalize_vector(residual)

    @staticmethod
    def goal(target, prediction):
        """
        """
        return normalize_vector(target)



from jax_fdm.geometry import closest_point_on_segment

from jax_fdm.goals.node import NodeLineGoal


class NodeSegmentGoal(NodeLineGoal):
    """
    Pulls the xyz position of a node to a target line ray.
    """
    @staticmethod
    def goal(target, prediction):
        """
        The closes point on the target seegment.
        """
        return closest_point_on_segment(prediction, target)



import jax.numpy as jnp


class Regularizer:
    def __init__(self, alpha, name=None):
        self.alpha = alpha
        self.name = name or self.__class__.__name__


class L2Regularizer(Regularizer):
    def __call__(self, params):
        # TODO: Apply jnp.sqrt to sum of squares to make it a valid L2 norm?
        return self.alpha * jnp.sum(jnp.square(params.q))



import numpy as np

from jax import vmap

import jax.numpy as jnp


class Constraint:
    """
    Base class for all constraints.
    """
    def __init__(self, key, bound_low=None, bound_up=None):
        self._key = None
        self.key = key

        self._bound_low = None
        self.bound_low = bound_low

        self._bound_up = None
        self.bound_up = bound_up

        self._index = None

    @property
    def key(self):
        """
        The key of an element in a network.
        """
        return self._key

    @key.setter
    def key(self, key):
        self._key = key

    @property
    def index(self):
        """
        The index of the goal key in the canonical ordering of an equilibrium structure.
        """
        return self._index

    @index.setter
    def index(self, index):
        if isinstance(index, int):
            index = [index]
        self._index = np.array(index)

    @staticmethod
    def _bound_setter(bound):
        """
        Set a bound.
        """
        if not isinstance(bound, (int, float)):
            if len(bound) == 1:
                bound = bound[0]
            else:
                bound = np.ravel(bound)
        return bound

    @property
    def bound_low(self):
        """
        The lower bound of this constraint.
        """
        return self._bound_low

    @bound_low.setter
    def bound_low(self, bound):
        if bound is None:
            bound = -jnp.inf
        self._bound_low = self._bound_setter(bound)

    @property
    def bound_up(self):
        """
        The upper bound of this constraint.
        """
        return self._bound_up

    @bound_up.setter
    def bound_up(self, bound):
        if bound is None:
            bound = jnp.inf
        self._bound_up = self._bound_setter(bound)

    @staticmethod
    def index_from_model(model):
        """
        Get the index in the model of the constraint key.
        """
        raise NotImplementedError

    def init(self, model, structure):
        """
        Initialize the constraint with information from an equilibrium model.
        """
        self.index = self.index_from_model(model, structure)

    def __call__(self, params, model, structure):
        """
        The called constraint function.
        """
        eqstate = model(params, structure)
        constraint = vmap(self.constraint, in_axes=(None, 0))(eqstate, self.index)
        # assert jnp.ravel(constraint).shape == jnp.ravel(self.index).shape, f"Constraint shape: {constraint.shape} vs. index shape: {self.index.shape}"

        return jnp.ravel(constraint)

    def constraint(self, eqstate, index):
        """
        The constraint function.
        """
        raise NotImplementedError



from jax_fdm.goals import Goal


class NodeGoal(Goal):
    """
    Base class for all goals that pertain to the node of a network.
    """
    def __init__(self, key, target, weight=1.0):
        super().__init__(key=key, target=target, weight=weight)

    def index_from_model(self, model, structure):
        """
        The index of the edge in a structure.
        """
        try:
            return structure.node_index[self.key]
        except TypeError:
            return tuple([structure.node_index[k] for k in self.key])



import jax.numpy as jnp


# ==========================================================================
# Mixins
# ==========================================================================

class IndexingMixins:
    @property
    def node_index(self):
        """
        A dictionary between node keys and their enumeration indices.
        """
        return {int(node): index for index, node in enumerate(self.nodes)}

    @property
    def edge_index(self):
        """
        A dictionary between edge keys and their enumeration indices.
        """
        return {(int(u), int(v)): index for index, (u, v) in enumerate(self.edges)}

    def _edges_indexed(self):
        """
        An iterable with the edges pointing to the indices of the node keys.
        """
        node_index = self.node_index

        edges_indexed = []
        for u, v in self.edges:
            edge = node_index[int(u)], node_index[int(v)]
            edges_indexed.append(edge)

        return jnp.array(edges_indexed)


class MeshIndexingMixins:
    @property
    def vertex_index(self):
        """
        A dictionary between vertex keys and their enumeration indices.
        """
        return {int(node): index for index, node in enumerate(self.vertices)}

    def _faces_indexed(self):
        """
        An array of the faces pointing to the indices of the node keys.
        """
        vertex_index = self.vertex_index

        findexed = []
        for face in self.faces:
            face_indices = []

            for vertex in face:
                u = int(vertex)
                if u >= 0:
                    index = vertex_index[u]
                else:
                    index = u
                face_indices.append(index)

            findexed.append(tuple(face_indices))

        return jnp.array(findexed)



from jax.numpy import inf


# ==========================================================================
# Parameter
# ==========================================================================

class Parameter:
    """
    The base class for all parameters.

    Parameters
    ----------
    key : `int` | `Tuple[int]`
        The key in the network of the parameter.
    bound_low : `float`, optional
        The lower bound of this parameter for optimization.
        Defaults to `-inf`.
    bound_ip : `float`, optional
        The upper bound of this parameter for optimization.
        Defaults to `+inf`.
    """
    def __init__(self, key, bound_low=None, bound_up=None):
        """
        Initialize the parameter.
        """
        self.key = key
        self.attr_name = None

        self._bound_low = None
        self.bound_low = bound_low

        self._bound_up = None
        self.bound_up = bound_up

    @property
    def bound_low(self):
        """
        The lower bound of the parameter.
        """
        return self._bound_low

    @property
    def bound_up(self):
        """
        The upper bound of the parameter.
        """
        return self._bound_up

    @bound_low.setter
    def bound_low(self, value):
        if value is None:
            value = -inf
        self._bound_low = value

    @bound_up.setter
    def bound_up(self, value):
        if value is None:
            value = inf
        self._bound_up = value

    def index(self, model):
        """
        Get the index of the parameter key in the structure of a model.
        """
        raise NotImplementedError

    def value(self, model, network):
        """
        Get the current value of a paramter from the structure of a model.
        """
        raise NotImplementedError


# ==========================================================================
# Individual parameters
# ==========================================================================

class NodeParameter(Parameter):
    """
    A node parameter.
    """
    def index(self, model, structure):
        """
        Get the index of the key of the node parameter in the structure of a model.
        """
        return structure.node_index[self.key]

    def value(self, model, network):
        """
        Get the current value of the node parameter.
        """
        return network.node_attribute(self.key, name=self.attr_name)


class VertexParameter(Parameter):
    """
    A vertex parameter.
    """
    def index(self, model, structure):
        """
        Get the index of the key of the node parameter in the structure of a model.
        """
        return structure.vertex_index[self.key]

    def value(self, model, mesh):
        """
        Get the current value of the node parameter.
        """
        return mesh.vertex_attribute(self.key, name=self.attr_name)


class EdgeParameter(Parameter):
    """
    An edge parameter.
    """
    def index(self, model, structure):
        """
        Get the index of the key of the edge parameter in the structure of a model.
        """
        return structure.edge_index[self.key]

    def value(self, model, datastructure):
        """
        Get the current value of the edge parameter.
        """
        return datastructure.edge_attribute(self.key, name=self.attr_name)


# ==========================================================================
# Parameter groups
# ==========================================================================

class ParameterGroup(Parameter):
    """
    A parent class for groups of parameters.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        assert len(self.key) > 0


class NodeGroupParameter(ParameterGroup):
    """
    A single parameter applied to a group of nodes.
    """
    def index(self, model, structure):
        """
        Get the indices of the keys of the parametrized nodes from the structure of a model.
        """
        return [structure.node_index[key] for key in self.key]

    def value(self, model, network):
        """
        Get the current average value of the parameter of the grouped nodes.
        """
        values = [network.node_attribute(key, name=self.attr_name) for key in self.key]
        return sum(values) / len(values)


class VertexGroupParameter(ParameterGroup):
    """
    A single parameter applied to a group of nodes.
    """
    def index(self, model, structure):
        """
        Get the indices of the keys of the parametrized vertices of a structure.
        """
        return [structure.vertex_index[key] for key in self.key]

    def value(self, model, mesh):
        """
        Get the current average value of the parameter of the grouped vertices.
        """
        values = [mesh.vertex_attribute(key, name=self.attr_name) for key in self.key]
        return sum(values) / len(values)


class EdgeGroupParameter(ParameterGroup):
    """
    A single parameter applied to a group of edges.
    """
    def index(self, model, structure):
        """
        Get the indices of the keys of the parametrized edges from the structure of a model.
        """
        return [structure.edge_index[key] for key in self.key]

    def value(self, model, datastructure):
        """
        Get the current average value of the parameter of the grouped edges.
        """
        values = [datastructure.edge_attribute(key, name=self.attr_name) for key in self.key]
        return sum(values) / len(values)


# ==========================================================================
# Edge force densities
# ==========================================================================

class EdgeForceDensityParameter(EdgeParameter):
    """
    An edge force density parameter.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.attr_name = "q"


class EdgeGroupForceDensityParameter(EdgeGroupParameter, EdgeForceDensityParameter):
    """
    A single force density value to rule them all.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


# ==========================================================================
# Node supports
# ==========================================================================

class NodeSupportParameter(NodeParameter):
    """
    A node support parameter.
    """
    def index(self, model, structure):
        """
        Get the index of the key of the node support in the structure of a model.
        """
        return structure.support_index[self.key]


class NodeSupportXParameter(NodeSupportParameter):
    """
    Parametrize the X coordinate of a support node.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.attr_name = "x"


class NodeSupportYParameter(NodeSupportParameter):
    """
    Parametrize the Y coordinate of a support node.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.attr_name = "y"


class NodeSupportZParameter(NodeSupportParameter):
    """
    Parametrize the Z coordinate of a support node.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.attr_name = "z"


# ==========================================================================
# Node supports groups
# ==========================================================================


class NodeGroupSupportParameter(NodeGroupParameter):
    """
    Parametrize a group of support nodes.
    """
    def index(self, model, structure):
        """
        Get the indices of the keys of the parametrized nodes from the structure of a model.
        """
        return [structure.support_index[key] for key in self.key]


class NodeGroupSupportXParameter(NodeGroupSupportParameter, NodeSupportXParameter):
    """
    Parametrize wiht a single value the X coordinate of a group of support nodes.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class NodeGroupSupportYParameter(NodeGroupSupportParameter, NodeSupportYParameter):
    """
    Parametrize wiht a single value the Y coordinate of a group of support nodes.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class NodeGroupSupportZParameter(NodeGroupSupportParameter, NodeSupportZParameter):
    """
    Parametrize wiht a single value the Z coordinate of a group of support nodes.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


# ==========================================================================
# Node loads
# ==========================================================================

class NodeLoadParameter(NodeParameter):
    pass


class NodeLoadXParameter(NodeLoadParameter):
    """
    Parametrize the x component of a node load.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.attr_name = "px"


class NodeLoadYParameter(NodeLoadParameter):
    """
    Parametrize the y component of a node load.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.attr_name = "py"


class NodeLoadZParameter(NodeLoadParameter):
    """
    Parametrize the z component of a node load.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.attr_name = "pz"


# ==========================================================================
# Node loads groups
# ==========================================================================

class NodeGroupLoadXParameter(NodeGroupParameter, NodeLoadXParameter):
    """
    Parametrize wiht a single value the X component of the load applied to a group of nodes.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class NodeGroupLoadYParameter(NodeGroupParameter, NodeLoadXParameter):
    """
    Parametrize wiht a single value the Y component of the load applied to a group of nodes.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class NodeGroupLoadZParameter(NodeGroupParameter, NodeLoadZParameter):
    """
    Parametrize wiht a single value the Z component of the load applied to a group of nodes.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


# ==========================================================================
# Vertex supports
# ==========================================================================

class VertexSupportParameter(VertexParameter):
    """
    A node support parameter.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class VertexSupportXParameter(VertexSupportParameter, NodeSupportXParameter):
    """
    Parametrize the X coordinate of a support node.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class VertexSupportYParameter(VertexSupportParameter, NodeSupportYParameter):
    """
    Parametrize the Y coordinate of a support node.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class VertexSupportZParameter(VertexSupportParameter, NodeSupportZParameter):
    """
    Parametrize the Z coordinate of a support node.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


# ==========================================================================
# Vertex supports groups
# ==========================================================================


class VertexGroupSupportParameter(NodeGroupParameter):
    """
    Parametrize a group of support nodes.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class VertexGroupSupportXParameter(VertexGroupParameter, VertexSupportXParameter):
    """
    Parametrize wiht a single value the X coordinate of a group of support nodes.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class VertexGroupSupportYParameter(VertexGroupParameter, VertexSupportYParameter):
    """
    Parametrize wiht a single value the Y coordinate of a group of support nodes.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class VertexGroupSupportZParameter(VertexGroupParameter, VertexSupportZParameter):
    """
    Parametrize wiht a single value the Z coordinate of a group of support nodes.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


# ==========================================================================
# Vertex loads
# ==========================================================================

class VertexLoadParameter(VertexParameter):
    pass


class VertexLoadXParameter(VertexLoadParameter, NodeLoadXParameter):
    """
    Parametrize the x component of a node load.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class VertexLoadYParameter(VertexLoadParameter, NodeLoadYParameter):
    """
    Parametrize the y component of a node load.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class VertexLoadZParameter(VertexLoadParameter, NodeLoadZParameter):
    """
    Parametrize the z component of a node load.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


# ==========================================================================
# Vertex loads groups
# ==========================================================================

class VertexGroupLoadXParameter(VertexGroupParameter, VertexLoadXParameter):
    """
    Parametrize wiht a single value the X component of the load applied to a group of nodes.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class VertesGroupLoadYParameter(VertexGroupParameter, VertexLoadXParameter):
    """
    Parametrize wiht a single value the Y component of the load applied to a group of nodes.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class VertexGroupLoadZParameter(VertexGroupParameter, VertexLoadZParameter):
    """
    Parametrize wiht a single value the Z component of the load applied to a group of nodes.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



import jax.numpy as jnp

from jax import vmap

from jax_fdm.geometry import area_triangle
from jax_fdm.geometry import normal_polygon
from jax_fdm.geometry import length_vector
from jax_fdm.geometry import line_lcs
from jax_fdm.geometry import polygon_lcs


# ==========================================================================
# Face loads
# ==========================================================================

from jax_fdm.geometry import normalize_vector


def nodes_load_from_faces(xyz, faces_load, structure, is_local=False):
    """
    Calculate the tributary face loads aplied to the nodes of a structure.
    """
    faces = structure.faces_indexed
    faces_load = calculate_faces_load(xyz, faces, faces_load, is_local)
    edges_load = edges_tributary_faces_load(xyz, faces_load, structure)
    nodes_load = nodes_tributary_edges_load(edges_load, structure)

    return nodes_load


def calculate_faces_load(xyz, faces, faces_load, is_local):
    """
    Transform the face loads to the XYZ cartesian coordinate system if needed.
    """
    if not is_local:
        return faces_load

    faces_load_lcs = vmap(face_load_lcs, in_axes=(None, 0, 0))
    loads = faces_load_lcs(xyz, faces, faces_load)

    return loads


def face_load_lcs(xyz, face, face_load):
    """
    Transform the load vector applied to the face to a vector in its local coordinate system.
    """
    def face_xyz(xyz, face):
        """
        Get this face XYZ coordinates from XYZ vertices array.
        """
        face = jnp.ravel(face)

        xyz_face = xyz[face, :]
        xyz_repl = xyz_face[0, :]

        # NOTE: Replace -1 with first entry to avoid nans in gradient computation
        # This was a pesky bug, since using nans as replacement did not cause
        # issues with the forward computation of normals, but it does for
        # the backward pass.
        xyz_face = vmap(jnp.where, in_axes=(0, 0, None))(face >= 0, xyz_face, xyz_repl)

        return xyz_face

    fxyz = face_xyz(xyz, face)

    normal = normal_polygon(fxyz)
    is_zero_normal = jnp.allclose(normal, 0.0)
    lcs = jnp.where(is_zero_normal, jnp.eye(3), polygon_lcs(fxyz))

    load = face_load @ lcs

    return load


def edges_tributary_faces_load(xyz, faces_load, structure):
    """
    Calculate the face area load taken by every edge in a datastructure.
    """
    c_vertices = structure.connectivity
    c_faces = structure.connectivity_edges_faces

    face_centroids = structure.connectivity_faces_vertices @ xyz
    loads_fn = vmap(edge_tributary_faces_load, in_axes=(0, 0, None, None, None))

    return loads_fn(c_vertices, c_faces, xyz, faces_load, face_centroids)


def edge_tributary_faces_load(c_edge_nodes, c_edge_faces, xyz, faces_load, face_centroids):
    """
    Calculate the face area load taken by one edge in a datastructure.
    """
    indices = jnp.flatnonzero(c_edge_nodes, size=2)
    line = xyz[indices, :]

    findices = jnp.flatnonzero(c_edge_faces, size=2, fill_value=-1)
    floads = faces_load[findices, :]

    centroids = face_centroids[findices, :]

    # NOTE: correct loads if negative face indices exist (e.g. faces on boundary)
    areas = vmap(edge_tributary_face_area, in_axes=(None, 0))(line, centroids)
    areas = jnp.where(findices >= 0, areas.ravel(), 0.0)

    return areas @ floads


def edge_tributary_face_area(line, centroid):
    """
    The triangle-based, face tributary area of an edge.
    """
    triangle = jnp.vstack((line, jnp.reshape(centroid, (1, 3))))

    return area_triangle(triangle)


def _faces_load_2(xyz, faces, faces_load, is_local):
    """
    Transform the face loads to the XYZ cartesian coordinate system.
    """
    def _faces_load(xyz, face, face_load, is_local):
        return jnp.where(is_local, face_load_lcs(xyz, face, face_load), face_load)

    floads, is_local = faces_load
    vmap_facesload = vmap(_faces_load, in_axes=(None, 0, 0, 0))

    return vmap_facesload(xyz, faces, faces_load, is_local)


# ==========================================================================
# Edge loads
# ==========================================================================


def nodes_load_from_edges(xyz, edges_load, structure, is_local=False):
    """
    Calculate the tributary edge loads aplied to the nodes of a structure.
    """
    edges = structure.edges_indexed
    edges_load = calculate_edges_load(xyz, edges, edges_load, is_local)
    edges_load = edges_tributary_edges_load(xyz, edges_load, structure)

    return nodes_tributary_edges_load(edges_load, structure)


def calculate_edges_load(xyz, edges, edges_load, is_local):
    """
    Transform the edges load to the XYZ cartesian coordinate system if needed.
    """
    if not is_local:
        return edges_load

    edges_load_lcs = vmap(edge_load_lcs, in_axes=(None, 0, 0))

    return edges_load_lcs(xyz, edges, edges_load)


def edge_load_lcs(xyz, edge, edge_load):
    """
    Transform the load vector applied to an edge to a vector in its local coordinate system.
    """
    def edge_xyz(xyz, edge):
        return xyz[edge, :]

    exyz = edge_xyz(xyz, edge)
    lcs = line_lcs(exyz)

    return edge_load @ lcs


def edges_tributary_edges_load(xyz, edges_load, structure):
    """
    Calculate the face area load taken by every edge in a datastructure.
    """
    # TODO: edge lengths calculated by the FDM equilibrium model, inject?
    edges_vector = structure.connectivity @ xyz
    edges_length = length_vector(edges_vector)

    # scale edge load by edge length
    return edges_load * edges_length


# ==========================================================================
# Node helpers
# ==========================================================================

def nodes_tributary_edges_load(edges_load, structure):
    """
    Calculate the load vector applied to the nodes based on the edge loads.
    """
    return 0.5 * jnp.abs(structure.connectivity).T @ edges_load


# ==========================================================================
# Main
# ==========================================================================

if __name__ == "__main__":

    import os
    from jax_fdm import DATA
    from jax_fdm.datastructures import FDMesh
    from jax_fdm.equilibrium import EquilibriumMeshStructure
    from jax_fdm.equilibrium import EquilibriumMeshStructureSparse

    # define tests
    def test_faces_load(mesh, pz, dtol=1e-1, sparse=False, is_local=False, verbose=False):

        mesh.faces_loads([0.0, 0.0, pz])

        if sparse:
            structure = EquilibriumMeshStructureSparse.from_mesh(mesh)
        else:
            structure = EquilibriumMeshStructure.from_mesh(mesh)

        xyz = jnp.array([mesh.vertex_coordinates(vkey) for vkey in mesh.vertices()])
        faces_load = jnp.asarray(mesh.faces_loads())

        faces = structure.faces_indexed
        nodes_load = nodes_load_from_faces(xyz, faces_load, structure)

        faces_load = calculate_faces_load(xyz, faces, faces_load, is_local)
        edges_load = edges_tributary_faces_load(xyz, faces_load, structure)

        sum_nodes_load = jnp.sum(nodes_load, axis=0)
        sum_edges_load = jnp.sum(edges_load, axis=0)

        compas_load = jnp.zeros(3)
        compas_faces_loads = []
        compas_faces_areas = []
        for face in mesh.faces():
            face_area = mesh.face_area(face)
            face_load = jnp.array(mesh.face_load(face)) * face_area
            compas_faces_areas.append(face_area)
            if is_local:
                fxyz = jnp.array(mesh.face_coordinates(face))
                face_load = face_load @ polygon_lcs(fxyz)
            compas_faces_loads.append(face_load)
            compas_load = compas_load + face_load

        if verbose:
            print(f"Face test: {sum_nodes_load=}, {sum_edges_load=}, {compas_load=:}")
            print("num faces", mesh.number_of_faces())
            print("num edges", mesh.number_of_edges())
            print("num nodes", mesh.number_of_vertices())
            print("edges load shape", edges_load.shape)
            print("nodes load shape", nodes_load.shape)
            print("faces load shape", faces_load.shape)

        # Test if resultant mesh loads match
        assert jnp.allclose(sum_edges_load, compas_load)
        assert jnp.allclose(sum_nodes_load, compas_load)

        # print(f"{compas_faces_loads=}")
        # Compare node loads individually
        compas_nodes_load = []
        face_index = {face: idx for idx, face in enumerate(mesh.faces())}

        if is_local:
            for vertex in structure.vertices:

                vertex = int(vertex)
                vertex_faceloads = [compas_faces_loads[face_index[face]] for face in mesh.vertex_faces(vertex)]
                vertex_faceareas = [compas_faces_areas[face_index[face]] for face in mesh.vertex_faces(vertex)]

                vertex_faceloads_unit = jnp.array(vertex_faceloads) / jnp.reshape(jnp.array(vertex_faceareas), (-1, 1))
                vload = jnp.sum(vertex_faceloads_unit, axis=0)
                vload = normalize_vector(vload) * pz

                # vload = jnp.sum(jnp.array(vertex_faceloads), axis=0)
                # varea = jnp.sum(jnp.array(vertex_faceareas), axis=0)
                # vload = vload / varea
                # compas_vertex_area = mesh.vertex_area(vertex)
                vload_unit = normalize_vector(vload)
                vnormal = jnp.array(mesh.vertex_normal(vertex))

                distance = jnp.linalg.norm(vload_unit - vnormal)

                # NOTE: Only checking if the load vector is in the same direction as the vertex normal
                # This is not checking that the load magnitudes are equal
                assert distance <= dtol, f"Distance: {distance}. Not equal: JAX: {vload_unit} vs. COMPAS: {vnormal}"
                compas_nodes_load.append(vload)

                # vload_unit = normalize_vector(jnp.sum(jnp.array(vertex_faceloads), axis=0))
                # vload = vload_unit * pz
                # compas_nodes_load.append(vload)

        else:
            compas_nodes_load = []
            for vertex in structure.vertices:
                vertex = int(vertex)
                vload = mesh.vertex_area(vertex) * pz
                compas_nodes_load.append([0.0, 0.0, vload])

            compas_nodes_load = jnp.array(compas_nodes_load)
            assert jnp.allclose(nodes_load, compas_nodes_load), f"Not equal: JAX:\n{nodes_load} vs. COMPAS:\n{compas_nodes_load}"

        mesh.faces_loads([0.0, 0.0, 0.0])
        # exit

    def test_edges_load(mesh, pz, dtol=1e-1, sparse=False, is_local=False, verbose=False):
        mesh.edges_loads([0.0, 0.0, pz])

        if sparse:
            structure = EquilibriumMeshStructureSparse.from_mesh(mesh)
        else:
            structure = EquilibriumMeshStructure.from_mesh(mesh)

        xyz = jnp.array([mesh.vertex_coordinates(vkey) for vkey in mesh.vertices()])

        edges = structure.edges_indexed
        edges_load = jnp.asarray(mesh.edges_loads())

        nodes_load = nodes_load_from_edges(xyz, edges_load, structure, is_local)

        edges_load_eff = calculate_edges_load(xyz, edges, edges_load, is_local)
        edges_load_trib = edges_tributary_edges_load(xyz, edges_load_eff, structure)

        sum_nodes_load = jnp.sum(nodes_load, axis=0)
        sum_edges_load = jnp.sum(edges_load_trib, axis=0)

        compas_load = jnp.zeros(3)
        for u, v in mesh.edges():
            edge_length = mesh.edge_length(u, v)
            edge_load = jnp.array(mesh.edge_load((u, v))) * edge_length
            if is_local:
                exyz = jnp.array(mesh.edge_coordinates(u, v))
                edge_load = edge_load @ line_lcs(exyz)
            compas_load = compas_load + edge_load

        if verbose:
            print(f"Edge test: {sum_nodes_load=}, {sum_edges_load=}, {compas_load=}")
            print("num edges", mesh.number_of_edges())
            print("num nodes", mesh.number_of_vertices())
            print("edges load shape", edges_load.shape)
            print("nodes load shape", nodes_load.shape)

        # Test if resultant mesh loads match
        assert jnp.allclose(sum_edges_load, compas_load)
        assert jnp.allclose(sum_nodes_load, compas_load)

    def test_mesh_loads(mesh, pz, dtol, sparse, is_local, verbose):
        test_faces_load(mesh, pz, dtol=dtol, sparse=sparse, is_local=is_local)
        test_edges_load(mesh, pz, dtol=dtol, sparse=sparse, is_local=is_local, verbose=verbose)

    # run tests
    pz = 2.0
    sparse = True
    is_local = False
    dtol = 1e-1
    verbose = False

    print("Test mesh grid")
    # for nx in range(1, 2):
    for nx in range(5, 10):
        mesh = FDMesh.from_meshgrid(dx=2, nx=nx)
        test_mesh_loads(mesh, pz, dtol=dtol, sparse=sparse, is_local=is_local, verbose=verbose)
    print()

    print("Test mesh polyhedron")
    for f in (4, 6, 8, 12, 20):
        mesh = FDMesh.from_polyhedron(f)
        test_mesh_loads(mesh, pz, dtol=dtol, sparse=sparse, is_local=is_local, verbose=verbose)
    print()

    print("Test mesh pillow")
    for name in ("uncstr_opt", "cstr_opt", "free", "loaded"):
        filepath = os.path.join(DATA, f"json/pillow_{name}_mesh.json")
        mesh = FDMesh.from_json(filepath)
        test_mesh_loads(mesh, pz, dtol=dtol, sparse=sparse, is_local=is_local, verbose=verbose)

    print("\nVery good! Slow down, cowboy!")
